"""
    middleware.deserializer
    ~~~~~~~~~~~~~~~~~~~~~~~

    Dynamically select & initialize the deserializer on the request
    object. The deserializer is chosen based on the Content-Type
    header provided.

    The deserializer is responsible for all of the deserialization
    logic, rules, & RFC compliance.
"""

import goldman

from goldman.exceptions import (
    EmptyRequestBody,
    ContentTypeUnsupported,
)
from goldman.utils.error_helpers import abort


class Middleware(object):
    """ Deserializer middleware object """

    @staticmethod
    def _get_deserializer(mimetype):
        """ Return a deserializer based on the mimetype

        If a deserializer can't be determined by the mimetype then
        return None.

        :param mimetype:
            string mimetype
        :return:
            deserializer class object or None
        """

        for deserializer in goldman.DESERIALIZERS:
            if mimetype == deserializer.MIMETYPE:
                return deserializer
        return None

    # pylint: disable=unused-argument
    def process_resource(self, req, resp, resource):
        """ Process the request after routing.

        Deserializer selection needs a resource to determine which
        deserializers are allowed. If a deserializer is required then
        it will be initialized & added to the request object for
        further processing.
        """

        if req.content_required and resource:
            allowed = resource.deserializer_mimetypes

            if req.content_length in (None, 0):
                abort(EmptyRequestBody)
            elif req.content_type not in allowed:
                abort(ContentTypeUnsupported(allowed))
            else:
                deserializer = self._get_deserializer(req.content_type)
                req.deserializer = deserializer(req, resp)



"""
    deserializers.form_data
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Deserializer that is compliant with the RFC 2388
    multipart/form-data format

    It's broken down into 2 parts:

        * A parser for spec compliant validations

        * A normalizer for converting into a common
          format expected by our resources/responders.


    WARN: Currently we only accept 1 upload at a time!
          An error message will be sent if more than 1
          upload is detected.
"""

import cgi
import goldman
import goldman.exceptions as exceptions
import goldman.extensions as extensions

from ..deserializers.base import Deserializer as BaseDeserializer
from goldman.utils.error_helpers import abort


class Deserializer(BaseDeserializer):
    """ RFC 2388 compliant deserializer """

    MIMETYPE = goldman.FILEUPLOAD_MIMETYPE

    def deserialize(self, mimetypes):  # pylint: disable=arguments-differ
        """ Invoke the deserializer

        Upon successful deserialization a dict will be returned
        containing the following key/vals:

            {
                'content': <uploaded object>,
                'content-type': <content-type of content>,
                'file-ext': <file extension based on content-type>,
                'file-name': <file name of content>,
            }

        :param mimetypes:
            allowed mimetypes of the object in the request
            payload
        :return:
            normalized dict
        """

        super(Deserializer, self).deserialize()

        parts = self.parse(mimetypes)
        data = self.normalize(parts)

        return data

    def normalize(self, parts):
        """ Invoke the RFC 2388 spec compliant normalizer

        :param parts:
            the already vetted & parsed FieldStorage objects
        :return:
            normalized dict
        """

        part = parts.list[0]

        return {
            'content': part.file.read(),
            'content-type': part.type,
            'file-ext': extensions.get(part.type),
            'file-name': part.filename,
        }

    def _parse_top_level_content_type(self):
        """ Ensure a boundary is present in the Content-Type header

        This is the Content-Type header outside of any form-data
        & should simply be:

            Content-Type: multipart/form-data; boundary=<value>\r\n

        This is generated by the client obviously & should not
        occur within the uploaded payload.
        """

        if not self.req.content_type_params.get('boundary'):
            abort(exceptions.InvalidRequestHeader(**{
                'detail': 'A boundary param is required in the Content-Type '
                          'header & cannot be an empty string. The details '
                          'of its grammar are further outlined in RFC 2046 '
                          '- section 5.1.1.',
                'links': 'tools.ietf.org/html/rfc2388#section-4.1',
            }))

    def _parse_section_three(self, part, mimetypes):
        """ Parse & validate a part according to section #3

        The logic applied follows section 3 guidelines from top
        to bottom.
        """

        link = 'tools.ietf.org/html/rfc2388#section-3'

        if part.disposition != 'form-data' or not part.name:
            self.fail('Each part of a multipart/form-data requires a '
                      'Content-Disposition header with a disposition type '
                      'of "form-data" AND a unique "name" parameter.', link)
        elif part.type.lower() not in mimetypes:
            allowed = ', '.join(mimetypes)
            self.fail('Invalid upload Content-Type. Each part of the '
                      'multipart/form-data upload MUST be one of: %s. '
                      '%s is not allowed.' % (allowed, part.type), link)

    def _parse_part(self, part, mimetypes):
        """ Validate each part of the multipart per RFC 2388 ""

        :param part:
            a FieldStorage object
        """

        self._parse_section_three(part, mimetypes)

    def parse(self, mimetypes):
        """ Invoke the RFC 2388 spec compliant parser """

        self._parse_top_level_content_type()

        link = 'tools.ietf.org/html/rfc2388'
        parts = cgi.FieldStorage(
            fp=self.req.stream,
            environ=self.req.env,
        )

        if not parts:
            self.fail('A payload in the body of your request is required '
                      '& must be encapsulated by the boundary with proper '
                      'headers according to RFC 2388', link)
        elif len(parts) > 1:
            self.fail('Currently, only 1 upload at a time is allowed. Please '
                      'break up your request into %s individual requests & '
                      'retry' % len(parts), link)
        else:
            self._parse_part(parts.list[0], mimetypes)
            return parts



"""
    serializers.csv
    ~~~~~~~~~~~~~~~

    Serializer that is compliant with the RFC 4180 CSV format
"""

import goldman

from ..serializers.base import Serializer as BaseSerializer


class Serializer(BaseSerializer):
    """ CSV serializer """

    MIMETYPE = goldman.CSV_MIMETYPE

    def serialize(self, data):
        """ Determine & invoke the proper serializer method

        If data is a list then the serialize_datas method will
        be run otherwise serialize_data.
        """

        super(Serializer, self).serialize(data)

        self.resp.content_type += '; header=present'
        import pprint
        pprint.PrettyPrinter().pprint(data)

        # X-Total-Count header
        # Link header for pagination
        #   tools.ietf.org/html/rfc5988#section-5
        #   canvas.instructure.com/doc/api/file.pagination.html
        #   github.com/falconry/falcon/blob/master/falcon/response.py#L379

    def _serialize_datas(self, datas):
        """ Turn the list into individual CSV records (rows)

        :spec: jsonapi.org/format/#document-top-level
        :param datas: list of dicts
        :return: list
        """

        return [self._serialize_data(d) for d in datas]

    def _serialize_data(self, data):
        """ Turn the data into a single CSV record (row)

        :spec: jsonapi.org/format/#document-resource-objects
        :param data: dict
        :return: dict
        """

        data.update(data['include'])
        data.update(data['to_many'])
        data.update(data['to_one'])

        for key, val in data.items():
            if isinstance(val, dict):
                pass
            elif isinstance(val, list):
                pass
            else:
                str(val)



"""
    middleware.model_qps
    ~~~~~~~~~~~~~~~~~~~~

    This middleware will initialize the query parameters supported
    by our model based resources. The supported query params are
    those directly called out by the JSON API specification but
    do not require the JSON API (de)serializer.


    NOTE: Query param processing will NOT occur on resources
          without a model property containing a goldman model
          class object. The model is needed so proper vetting
          of the query params can occur against the models
          attributes.


    The request object will have it's attributes updated with
    the coerced query parameter values. The attribute names are
    the pluralized version of the query params themselves.

    Currently supported query params are:

        fields, filter, include, page, & sort
"""

import goldman.queryparams.fields as fields
import goldman.queryparams.filter as filters
import goldman.queryparams.include as includes
import goldman.queryparams.page as pages
import goldman.queryparams.sort as sorts

from goldman.utils.error_helpers import abort
from goldman.exceptions import InvalidQueryParams


class Middleware(object):
    """ Model query parameter middleware. """

    # pylint: disable=unused-argument
    def process_resource(self, req, resp, resource):
        """ Process the request after routing.

        The resource is required to determine if a model based
        resource is being used for validations so skip
        processing if no `resource.model` attribute is present.
        """

        try:
            model = resource.model

            req.fields = fields.init(req, model)
            req.filters = filters.init(req, model)
            req.includes = includes.init(req, model)
            req.pages = pages.init(req, model)
            req.sorts = sorts.init(req, model)
        except AttributeError:
            pass
        except InvalidQueryParams as exc:
            abort(exc)



"""
    config
    ~~~~~~

    Initialize some of our helpful constants used throughout
    the api & merge them with any app specific overrides.
"""

import importlib
import os


class Config(object):
    """ Default configuration settings.

    To override any these configurations define the env
    variable 'GOLDMAN_CONFIG_MODULE' & ensure the module
    is in the python path.

    Any directives beginning with '_' will be considered
    private & not attached to the Config object.
    """

    # JSON API
    JSONAPI_VERSION = '1.0'

    # Query filter operators
    BOOL_FILTERS = ('exists',)
    DATE_FILTERS = ('after', 'before')
    ENUM_FILTERS = ('in', 'len', 'nin')
    EQUAL_FILTERS = ('eq', 'neq')
    GEO_FILTERS = ('near', 'within')
    NUM_FILTERS = ('gt', 'gte', 'lt', 'lte')
    STR_FILTERS = ('contains', 'icontains', 'endswith', 'startswith')

    QUERY_FILTERS = BOOL_FILTERS + DATE_FILTERS + ENUM_FILTERS + \
        EQUAL_FILTERS + GEO_FILTERS + NUM_FILTERS + STR_FILTERS

    # Query pagination
    PAGE_LIMIT = 10

    # Query sort preference
    SORT = 'created'

    # Security stuff
    MAX_URI_LENGTH = 4000
    TLS_REQUIRED = True

    # URL prefix for the API
    BASE_URL = '/api'

    # WWW-Authenticate
    AUTH_REALM = 'JSON API'

    def __init__(self):

        try:
            module = os.environ.get('GOLDMAN_CONFIG_MODULE', '')
            module = importlib.import_module(module)

            for attr in dir(module):
                if not attr.startswith('_'):
                    val = getattr(module, attr)
                    setattr(self, attr, val)
        except (ImportError, ValueError):
            module = None

    def __getattr__(self, name):
        """ Return None if the attr isn't found

        Instead of throwing an exception if the attribute isn't
        found like python would ordinarily do & forcing each access
        to guard against... we conveniently return None.
        """

        try:
            return super(Config).__getattr__(name)
        except AttributeError:
            return None



"""
    resources.json_7159
    ~~~~~~~~~~~~~~~~~~~

    JSON based resource to be sub-classed with responders.
"""

import goldman

from ..resources.base import Resource as BaseResource


class Resource(BaseResource):
    """ Base JSON resource """

    DESERIALIZERS = [
        goldman.JsonDeserializer,
    ]

    SERIALIZERS = [
        goldman.JsonSerializer,
    ]



"""
    types.zipcode
    ~~~~~~~~~~~~~

    schematics zip code type
"""

import re

from schematics.exceptions import ValidationError
from schematics.types import StringType


class Type(StringType):
    """ U.S. zip code field with validation """

    MESSAGES = {
        'zip_code': 'Invalid zip code format'
    }

    ZIPCODE_REGEX = re.compile('^[0-9]{5}(?:-[0-9]{4})?$')

    def validate_zip_code(self, value):
        """ Schematics validator """

        if not Type.ZIPCODE_REGEX.match(value):
            raise ValidationError(self.messages['zip_code'])
        return value



"""
    request
    ~~~~~~~

    Define our sub-classed Request object to be used instead of the
    native falcon Request object.

    We extend the native Falcon request object in the following
    notable ways:

      * break up the Content-Type itself & its params into
        separate properties. We override the native content_type
        property to NOT include the params.

      * proxy the deserialize method to the selected
        deserializer

      * proxy the deserialize method to the selected
        deserializer

      * is_<method> helpers to evaluate the type of method
        being used as a truthy property
"""

from falcon.request import Request as FalconRequest
from goldman.utils.str_helpers import naked


class Request(FalconRequest):
    """ Subclass the default falcon request object

    First set content_type_params before reassigning content_type
    sans params.
    """

    def __init__(self, *args, **kwargs):

        super(Request, self).__init__(*args, **kwargs)

        self.content_type_params = self._init_content_type_params()
        self.content_type = self._init_content_type()
        self.deserializer = None

    @property
    def auth_scheme(self):
        """ If an Authorization header is present get the scheme

        It is expected to be the first string in a space separated
        list & will always be returned lowercase.

        :return: str or None
        """

        try:
            auth = getattr(self, 'auth')
            return naked(auth.split(' ')[0]).lower()
        except (AttributeError, IndexError):
            return None

    @property
    def content_required(self):
        """ Check if a Content-* request headers are needed """

        return self.method in ('PATCH', 'POST', 'PUT')

    @property
    def is_deleting(self):
        """ Return True if the request method is DELETE """

        return self.method == 'DELETE'

    @property
    def is_getting(self):
        """ Return True if the request method is GET """

        return self.method == 'GET'

    @property
    def is_patching(self):
        """ Return True if the request method is PATCH """

        return self.method == 'PATCH'

    @property
    def is_posting(self):
        """ Return True if the request method is POST """

        return self.method == 'POST'

    def _init_content_type(self):
        """ Return the Content-Type request header excluding params

        Remove any excess whitespace & lower case it for more
        predictable compares.
        """

        try:
            return naked(self.content_type.split(';')[0]).lower()
        except (AttributeError, IndexError):
            return None

    def _init_content_type_params(self):
        """ Return the Content-Type request header parameters

        Convert all of the semi-colon separated parameters into
        a dict of key/vals. If for some stupid reason duplicate
        & conflicting params are present then the last one
        wins.

        If a particular content-type param is non-compliant
        by not being a simple key=val pair then it is skipped.

        If no content-type header or params are present then
        return an empty dict.

        :return: dict
        """

        ret = {}

        if self.content_type:
            params = self.content_type.split(';')[1:]
            for param in params:
                try:
                    key, val = param.split('=')
                    ret[naked(key)] = naked(val)
                except ValueError:
                    continue

        return ret

    def deserialize(self, *args, **kwargs):
        """ Simple proxy to the deserializer's deserialize function

        This allows code to later run request.deserialize() &
        have it always call the deserialize method on the proper
        deserializer.
        """

        return self.deserializer.deserialize(*args, **kwargs)

    def get_body(self):
        """ Read in the request stream & return it as is

        If the request object doesn't claim to have a request body
        eg no content_length, then return None

        WARN: This could use a ton of memory so use it wisely.
              Images or bulk payloads should be read in more
              intelligently by the deserializer.

        :return:
            request payload as is
        """

        if self.content_length in (None, 0):
            return None

        # pylint: disable=no-member
        return self.stream.read()



"""
    deserializers.jsonapi
    ~~~~~~~~~~~~~~~~~~~~~

    Deserializer that is compliant with the JSON API specification
    as documented here:

        http://jsonapi.org/

    It's broken down into 2 parts:

        1) A parser for spec compliant validations
        2) A normalizer for converting into a common
           format expected by our resources/responders.
"""

import goldman
import goldman.exceptions as exceptions

from ..deserializers.json_7159 import Deserializer as JsonDeserializer
from goldman.utils.error_helpers import abort


class Deserializer(JsonDeserializer):
    """ JSON API compliant deserializer

    Inherit from the JsonDeserializer since JSON API is
    simply a structured representation of JSON.
    """

    MIMETYPE = goldman.JSONAPI_MIMETYPE

    def deserialize(self):
        """ Invoke the deserializer

        The JsonDeserializer.deserializer() method will
        return a dict of the JSON payload.

        :return:
            normalized dict
        """

        body = super(Deserializer, self).deserialize()

        self.parse(body)
        return self.normalize(body)

    @staticmethod
    def _normalize_attributes(attributes):
        """ Get all the attributes by key/val & return them

        :param attributes:
            dict JSON API attributes object
        :return: dict
        """

        return attributes

    @staticmethod
    def _normalize_relationships(relationships):
        """ Get all the relationships by key/val & return them

        A normalized relationship dict uses the key name without
        any alteration but the value will be the `id` provided
        in the payload if present. If not present, then the client
        wants to unset the relationship so it will be set to None.

        INFO: only works for to-one relationships.

        :param relationships:
            dict JSON API relationships object
        :return: dict
        """

        ret = {}

        for key, val in relationships.items():
            if not val['data']:
                ret[key] = None
            else:
                ret[key] = val['data']['id']

        return ret

    def normalize(self, body):
        """ Invoke the JSON API normalizer

        Perform the following:

            * add the type as a rtype property
            * flatten the payload
            * add the id as a rid property ONLY if present

        We don't need to vet the inputs much because the Parser
        has already done all the work.

        :param body:
            the already vetted & parsed payload
        :return:
            normalized dict
        """

        resource = body['data']
        data = {'rtype': resource['type']}

        if 'attributes' in resource:
            attributes = resource['attributes']
            attributes = self._normalize_attributes(attributes)

            data.update(attributes)

        if 'relationships' in resource:
            relationships = resource['relationships']
            relationships = self._normalize_relationships(relationships)

            data.update(relationships)

        if resource.get('id'):
            data['rid'] = resource['id']

        return data

    def _parse_attributes(self, attributes):
        """ Ensure compliance with the spec's attributes section

        Specifically, the attributes object of the single resource
        object. This contains the key / values to be mapped to the
        model.

        :param attributes:
            dict JSON API attributes object
        """

        link = 'jsonapi.org/format/#document-resource-object-attributes'

        if not isinstance(attributes, dict):
            self.fail('The JSON API resource object attributes key MUST '
                      'be a hash.', link)
        elif 'id' in attributes or 'type' in attributes:
            self.fail('A field name of `id` or `type` is not allowed in '
                      'the attributes object. They should be top-level '
                      'keys.', link)

    def _parse_relationships(self, relationships):
        """ Ensure compliance with the spec's relationships section

        Specifically, the relationships object of the single resource
        object. For modifications we only support relationships via
        the `data` key referred to as Resource Linkage.

        :param relationships:
            dict JSON API relationships object
        """

        link = 'jsonapi.org/format/#document-resource-object-relationships'

        if not isinstance(relationships, dict):
            self.fail('The JSON API resource object relationships key MUST '
                      'be a hash & comply with the spec\'s resource linkage '
                      'section.', link)

        for key, val in relationships.items():
            if not isinstance(val, dict) or 'data' not in val:
                self.fail('Relationship key %s MUST be a hash & contain '
                          'a `data` field compliant with the spec\'s '
                          'resource linkage section.' % key, link)
            elif isinstance(val['data'], dict):
                data = val['data']
                rid = isinstance(data.get('id'), unicode)
                rtype = isinstance(data.get('type'), unicode)

                if not rid or not rtype:
                    self.fail('%s relationship\'s resource linkage MUST '
                              'contain `id` & `type` fields. Additionally, '
                              'they must both be strings.' % key, link)
            elif isinstance(val['data'], list):
                abort(exceptions.ModificationDenied(**{
                    'detail': 'Modifying the %s relationship or any to-many '
                              'relationships for that matter are is not '
                              'currently supported. Instead, modify the '
                              'to-one side directly.' % key,
                    'links': link,
                }))
            else:
                self.fail('The relationship key %s is malformed & impossible '
                          'for us to understand your intentions. It MUST be '
                          'a hash & contain a `data` field compliant with '
                          'the spec\'s resource linkage section or null if '
                          'you want to unset the relationship.' % key, link)

    def _parse_resource(self, resource):
        """ Ensure compliance with the spec's resource objects section

        :param resource:
            dict JSON API resource object
        """

        link = 'jsonapi.org/format/#document-resource-objects'
        rid = isinstance(resource.get('id'), unicode)
        rtype = isinstance(resource.get('type'), unicode)

        if not rtype or not (self.req.is_patching and not rid):
            self.fail('JSON API requires that every resource object MUST '
                      'contain a `type` top-level key. Additionally, when '
                      'modifying an existing resource object an `id` '
                      'top-level key is required. The values of both keys '
                      'MUST be strings. Your request did not comply with '
                      'one or more of these 3 rules', link)
        elif 'attributes' not in resource and 'relationships' not in resource:
            self.fail('Modifiying or creating resources require at minimum '
                      'an attributes object and/or relationship object.', link)
        elif rid and self.req.is_posting:
            abort(exceptions.ModificationDenied(**{
                'detail': 'Our API does not support client-generated ID\'s '
                          'when creating NEW resources. Instead, our API '
                          'will generate one for you & return it in the '
                          'response.',
                'links': 'jsonapi.org/format/#crud-creating-client-ids',
            }))

    def _parse_top_level(self, body):
        """ Ensure compliance with the spec's top-level section """

        link = 'jsonapi.org/format/#document-top-level'

        try:
            if not isinstance(body['data'], dict):
                raise TypeError
        except (KeyError, TypeError):
            self.fail('JSON API payloads MUST be a hash at the most '
                      'top-level; rooted at a key named `data` where the '
                      'value must be a hash. Currently, we only support '
                      'JSON API payloads that comply with the single '
                      'Resource Object section.', link)

        if 'errors' in body:
            self.fail('JSON API payloads MUST not have both `data` & '
                      '`errors` top-level keys.', link)

    def parse(self, body):
        """ Invoke the JSON API spec compliant parser

        Order is important. Start from the request body root key
        & work your way down so exception handling is easier to
        follow.

        :return:
            the parsed & vetted request body
        """

        self._parse_top_level(body)
        self._parse_resource(body['data'])

        resource = body['data']

        if 'attributes' in resource:
            self._parse_attributes(resource['attributes'])
        if 'relationships' in resource:
            self._parse_relationships(resource['relationships'])



"""
    middleware.serializer
    ~~~~~~~~~~~~~~~~~~~~~

    Dynamically select & initialize the serializer on the response
    object. The serializer is chosen based on the Accept header
    provided.

    The serializer is responsible for all of the serialization
    logic, rules, & RFC compliance.
"""

import goldman

from goldman.exceptions import RequestNotAcceptable
from goldman.utils.error_helpers import abort


class Middleware(object):
    """ Serializer middleware object """

    @staticmethod
    def _get_serializer(mimetype):
        """ Return a serializer based on the mimetype

        If a serializer can't be determined by the mimetype then
        return None.

        :param mimetype:
            string mimetype
        :return:
            serializer class object or None
        """

        for serializer in goldman.SERIALIZERS:
            if mimetype == serializer.MIMETYPE:
                return serializer
        return None

    def process_resource(self, req, resp, resource):
        """ Process the request after routing.

        Serializer selection needs a resource to determine which
        serializers are allowed.
        """

        if resource:
            allowed = resource.serializer_mimetypes
            preferred = req.client_prefers(allowed)

            if not preferred:
                abort(RequestNotAcceptable(allowed))
            else:
                serializer = self._get_serializer(preferred)
                resp.serializer = serializer(req, resp)



"""
    middleware.basicauth
    ~~~~~~~~~~~~~~~~~~~~

    Basic authentication implementation as documented briefly
    in RFC 7231 but more completely in RFC 2617 & 7235

        Per RFC 2617 section 2 an Authorization header in the
        request containing a base64 encoded, colon separated,
        username & password. It will come after the scheme
        is declared which is a string of "Basic".

    The middleware requires a callable to be passed in as the
    `auth_creds` property which will be given a username &
    password.

    The callable should return a login model of the successfully
    authenticated user or raise an `AuthRejected` exception
    causing the request to be aborted IF the middleware's
    `optional` property is set to False (default).

    The model will be assigned to the `goldman.sess.login`
    propery if authentication succeeds.
"""

import goldman
import goldman.signals as signals

from goldman.exceptions import (
    AuthRejected,
    AuthRequired,
    InvalidAuthSyntax,
)
from goldman.utils.error_helpers import abort
from goldman.utils.str_helpers import naked
from base64 import b64decode


class Middleware(object):
    """ Ensure RFC compliance & authenticate the user. """

    def __init__(self, auth_creds, optional=False):

        self.auth_creds = auth_creds
        self.optional = optional

    @property
    def _error_headers(self):
        """ Return a dict of headers in every auth failure """

        return {
            'Cache-Control': 'no-store',
            'Pragma': 'no-cache',
            'WWW-Authenticate': self._realm,
        }

    @property
    def _realm(self):
        """ Return a string representation of the authentication realm """

        return 'Basic realm="%s"' % goldman.config.AUTH_REALM

    def _validate_auth_scheme(self, req):
        """ Check if the request has auth & the proper scheme

        :raise:
            AuthRequired
        """

        if not req.auth:
            raise AuthRequired(**{
                'detail': 'You must first login to access the requested '
                          'resource(s). Please retry your request using '
                          'Basic Authentication as documented in RFC 2617 '
                          '& available in most HTTP client libraries.',
                'links': 'tools.ietf.org/html/rfc2617#section-2',
            })
        elif req.auth_scheme != 'basic':
            raise AuthRequired(**{
                'detail': 'Your Authorization header is using an unsupported '
                          'authentication scheme. Please modify your scheme '
                          'to be a string of: "Basic".',
                'links': 'tools.ietf.org/html/rfc2617#section-2',
            })

    def _get_creds(self, req):
        """ Get the username & password from the Authorization header

        If the header is actually malformed where Basic Auth was
        indicated by the request then an InvalidAuthSyntax exception
        is raised. Otherwise an AuthRequired exception since it's
        unclear in this scenario if the requestor was even aware
        Authentication was required & if so which "scheme".

        Calls _validate_auth_scheme first & bubbles up it's
        exceptions.

        :return:
            tuple (username, password)
        :raise:
            AuthRequired, InvalidAuthSyntax
        """

        self._validate_auth_scheme(req)

        try:
            creds = naked(req.auth.split(' ')[1])
            creds = b64decode(creds)

            username, password = creds.split(':')
            return username, password
        except IndexError:
            raise InvalidAuthSyntax(**{
                'detail': 'You are using the Basic Authentication scheme as '
                          'required to login but your Authorization header is '
                          'completely missing the login credentials.',
                'links': 'tools.ietf.org/html/rfc2617#section-2',
            })
        except TypeError:
            raise InvalidAuthSyntax(**{
                'detail': 'Our API failed to base64 decode your Basic '
                          'Authentication login credentials in the '
                          'Authorization header. They seem to be malformed.',
                'links': 'tools.ietf.org/html/rfc2617#section-2',
            })
        except ValueError:
            raise InvalidAuthSyntax(**{
                'detail': 'Our API failed to identify a username & password '
                          'in your Basic Authentication Authorization header '
                          'after decoding them. The username or password is '
                          'either missing or not separated by a ":" per the '
                          'spec. Either way the credentials are malformed.',
                'links': 'tools.ietf.org/html/rfc2617#section-2',
            })

    def process_request(self, req, resp):  # pylint: disable=unused-argument
        """ Process the request before routing it. """

        signals.pre_authenticate.send()

        try:
            creds = self._get_creds(req)
            goldman.sess.login = self.auth_creds(*creds)
            signals.post_authenticate.send()
        except (AuthRejected, AuthRequired, InvalidAuthSyntax) as exc:
            exc.headers = self._error_headers
            if not self.optional:
                abort(exc)



"""
    queryparams.fields
    ~~~~~~~~~~~~~~~~~~

    Limit the response to the requested sparse fields.

    The query parameter itself follows the JSON API convention
    for sparse fields:

        jsonapi.org/format/#fetching-sparse-fieldsets
        jsonapi.org/examples/#sparse-fieldsets
"""

import re

from goldman.exceptions import InvalidQueryParams
from goldman.utils.model_helpers import rtype_to_model


LINK = 'jsonapi.org/format/#fetching-sparse-fieldsets'
PARAM = 'fields'


def _parse_param(key, val):
    """ Parse the query param looking for sparse fields params

    Ensure the `val` or what will become the sparse fields
    is always an array. If the query param is not a sparse
    fields query param then return None.

    :param key:
        the query parameter key in the request (left of =)
    :param val:
        the query parameter val in the request (right of =)
    :return:
        tuple of resource type to implement the sparse
        fields on & a array of the fields.
    """

    regex = re.compile(r'fields\[([A-Za-z]+)\]')
    match = regex.match(key)

    if match:
        if not isinstance(val, list):
            val = val.split(',')

        fields = [field.lower() for field in val]
        rtype = match.groups()[0].lower()

        return rtype, fields


def _validate_param(rtype, fields):
    """ Ensure the sparse fields exists on the models """

    try:
        # raises ValueError if not found
        model = rtype_to_model(rtype)
        model_fields = model.all_fields
    except ValueError:
        raise InvalidQueryParams(**{
            'detail': 'The fields query param provided with a '
                      'field type of "%s" is unknown.' % rtype,
            'links': LINK,
            'parameter': PARAM,
        })

    for field in fields:
        if field not in model_fields:
            raise InvalidQueryParams(**{
                'detail': 'The fields query param "TYPE" of "%s" '
                          'is not possible. It does not have a field '
                          'by the name of "%s".' % (rtype, field),
                'links': LINK,
                'parameter': PARAM,
            })


def _validate_req(req):
    """ Ensure the request is a GET request

    Modification type requests could have other fields modified
    by the backend persistance layer through stored procedures or
    triggers & would need to be reported back to the client.

    In this scenario, it's best to only support sparse fields on
    requests that don't have side-effects.
    """

    if not req.is_getting:
        raise InvalidQueryParams(**{
            'detail': 'The fields query param is only supported '
                      'with GET requests.',
            'links': LINK,
            'parameter': PARAM,
        })


def init(req, model):  # pylint: disable=unused-argument
    """ Determine the sparse fields to limit the response to

    Return a dict where the key is the resource type (rtype) &
    the value is an array of string fields names to whitelist
    against.

    :return:
        dict
    """

    params = {}

    for key, val in req.params.items():
        try:
            rtype, fields = _parse_param(key, val)
            params[rtype] = fields
        except TypeError:
            continue

    if params:
        _validate_req(req)
        for rtype, fields in params.items():
            _validate_param(rtype, fields)
    return params



"""
    exceptions
    ~~~~~~~~~~

    Custom error classes to be raised throughout the app to
    abort requests with JSON API compliant errors
"""

import falcon

from goldman.utils.str_helpers import random_str


class APIException(Exception):
    """ Exception class for error handling in JSON API format.

    :spec: http://jsonapi.org/format/#errors
    """

    def __init__(self, **kwargs):

        super(APIException, self).__init__()

        kwargs['id'] = kwargs.get('id', random_str())
        kwargs['links'] = {'about': kwargs.get('links', '')}
        kwargs['detail'] = kwargs.get('detail', '')

        self.data = kwargs

    def __call__(self):

        return self

    def __getattr__(self, key):

        if key == "data":
            return self.data
        else:
            return self.data[key]

    def __setattr__(self, name, value):

        if name == 'data':
            super(APIException, self).__setattr__(name, value)
        else:
            self.data[name] = value

    def to_dict(self):
        """ Convenience function to get the exception as a dict """

        return self.data


"""
    400 Bad Request
    ~~~~~~~~~~~~~~~
"""


class EmptyRequestBody(APIException):
    """ The request body must not be empty """

    DETAIL = 'The operation you attempted requires a request body ' \
             'yet none was found. Please generate a valid request body ' \
             '& retry your request.'

    def __init__(self, **kwargs):
        super(EmptyRequestBody, self).__init__(**{
            'code': 'empty_request_body',
            'detail': self.DETAIL,
            'status': falcon.HTTP_400,
            'title': 'Empty request body',
        })


class InvalidQueryParams(APIException):
    """ The query parameter is invalid

    This exception supports detail & links overrides, however
    the source object containing a parameter to the attribute
    that caused the failure is required.
    """

    DETAIL = 'One or more query parameters are invalid or corrupt. ' \
             'Please modify your request & retry.'

    def __init__(self, **kwargs):
        super(InvalidQueryParams, self).__init__(**{
            'code': 'invalid_query_param',
            'detail': kwargs.get('detail', self.DETAIL),
            'links': kwargs.get('links'),
            'source': {'parameter': kwargs['parameter']},
            'status': falcon.HTTP_400,
            'title': 'Invalid or corrupt query parameter',
        })


class InvalidRequestBody(APIException):
    """ The request body is invalid

    This exception supports detail & links overrides.
    """

    DETAIL = 'The body of your request is fundamentally invalid or ' \
             'corrupt. We weren\'t able to even begin processing it ' \
             'because of the issue. Please modify your request body ' \
             '& retry.'

    def __init__(self, **kwargs):
        super(InvalidRequestBody, self).__init__(**{
            'code': 'invalid_request_body',
            'detail': kwargs.get('detail', self.DETAIL),
            'links': kwargs.get('links'),
            'status': falcon.HTTP_400,
            'title': 'Invalid or corrupt request body',
        })


class InvalidRequestHeader(APIException):
    """ One or more request headers are invalid

    This exception supports detail & links overrides.
    """

    DETAIL = 'One or more of your request headers and/or there ' \
             'parameters are invalid or corrupt. Please modify ' \
             'your request headers & retry.'

    def __init__(self, **kwargs):
        super(InvalidRequestHeader, self).__init__(**{
            'code': 'invalid_request_header',
            'detail': kwargs.get('detail', self.DETAIL),
            'links': kwargs.get('links'),
            'status': falcon.HTTP_400,
            'title': 'Invalid or corrupt request header',
        })


class InvalidURL(APIException):
    """ The URL is invalid

    This exception supports detail & links overrides.
    """

    DETAIL = 'The URL provided is invalid. Please modify your ' \
             'request URL & retry.'

    def __init__(self, **kwargs):
        super(InvalidURL, self).__init__(**{
            'code': 'invalid_url',
            'detail': kwargs.get('detail', self.DETAIL),
            'links': kwargs.get('links'),
            'status': falcon.HTTP_400,
            'title': 'Invalid or corrupt URL',
        })


"""
    401 Unauthorized
    ~~~~~~~~~~~~~~~~
"""


class AuthRejected(APIException):
    """ The auth provided was rejected for some reason. """

    DETAIL = 'The authentication credentials provided could not be ' \
             'validated. Double check your spelling & retry.'

    def __init__(self, **kwargs):
        super(AuthRejected, self).__init__(**{
            'code': 'auth_rejected',
            'detail': kwargs.get('detail', self.DETAIL),
            'headers': kwargs.get('headers'),
            'status': falcon.HTTP_401,
            'title': 'Authentication attempt rejected',
        })


class AuthRequired(APIException):
    """ The API requires some sort of Authentication

    This exception supports detail, headers, & links overrides.
    """

    DETAIL = 'The URL requested (route) requires authentication. ' \
             'Please retry your request with authentication.'

    def __init__(self, **kwargs):
        super(AuthRequired, self).__init__(**{
            'code': 'auth_required',
            'detail': kwargs.get('detail', self.DETAIL),
            'headers': kwargs.get('headers'),
            'links': kwargs.get('links'),
            'status': falcon.HTTP_401,
            'title': 'Authentication is required',
        })


class InvalidAuthSyntax(APIException):
    """ The Authentication provided is malformed. """

    DETAIL = 'The Authorization you provided could not be properly ' \
             'interpreted; it appears to be malformed. Please check ' \
             'your syntax.'

    def __init__(self, **kwargs):
        super(InvalidAuthSyntax, self).__init__(**{
            'code': 'invalid_auth_syntax',
            'detail': kwargs.get('detail', self.DETAIL),
            'headers': kwargs.get('headers'),
            'links': kwargs.get('links'),
            'status': falcon.HTTP_401,
            'title': 'Invalid authentication syntax',
        })


"""
    403 Forbidden
    ~~~~~~~~~~~~~
"""


class AccessDenied(APIException):
    """ The logged in user lacks privileges to get the resource. """

    DETAIL = 'The access attempt on the requested resource ' \
             'was denied.'

    def __init__(self, **kwargs):
        super(AccessDenied, self).__init__(**{
            'code': 'access_denied',
            'detail': kwargs.get('detail', self.DETAIL),
            'status': falcon.HTTP_403,
            'title': 'Access denied accessing resource',
        })


class ModificationDenied(APIException):
    """ The logged in user lacks privileges for attempted modification. """

    DETAIL = 'The modification attempt on the requested resource ' \
             'was denied.'

    def __init__(self, **kwargs):
        super(ModificationDenied, self).__init__(**{
            'code': 'modification_denied',
            'detail': kwargs.get('detail', self.DETAIL),
            'links': kwargs.get('links'),
            'status': falcon.HTTP_403,
            'title': 'Access denied in modification attempt',
        })


class TLSRequired(APIException):
    """ The API requires TLS (SSL) security """

    DETAIL = 'The URL requested (route) requires https (TLS). ' \
             'Any other protocols are not allowed. Please retry ' \
             'your request using https.'

    def __init__(self, **kwargs):
        super(TLSRequired, self).__init__(**{
            'code': 'ssl_required',
            'detail': self.DETAIL,
            'links': 'en.wikipedia.org/wiki/HTTPS',
            'status': falcon.HTTP_403,
            'title': 'TLS is required',
        })


"""
    404 Not Found
    ~~~~~~~~~~~~~
"""


class DocumentNotFound(APIException):
    """ The document queried for could not be found """

    DETAIL = 'We processed your request just fine but could not ' \
             'find the document you requested. Did it get deleted?'

    def __init__(self, **kwargs):

        super(DocumentNotFound, self).__init__(**{
            'code': 'document_not_found',
            'detail': self.DETAIL,
            'status': falcon.HTTP_404,
            'title': 'Queried document not found',
        })


class RouteNotFound(APIException):
    """ The API URL (route) requested does not exist """

    DETAIL = 'The URL requested (route) could not be found. ' \
             'We were unable to even query for any documents you may ' \
             'have been looking for because we have no routes registered ' \
             'by that URL. This is commonly a spelling error.'

    def __init__(self, **kwargs):

        super(RouteNotFound, self).__init__(**{
            'code': 'route_not_found',
            'detail': self.DETAIL,
            'status': falcon.HTTP_404,
            'title': 'API endpoint not found',
        })


"""
    405 Method Not Allowed
    ~~~~~~~~~~~~~~~~~~~~~~
"""


class MethodNotAllowed(APIException):
    """ The HTTP Method isn't allowed on the given endpoint (route) """

    DETAIL = 'The URL requested (route) does not support that method. ' \
             'We support a wide array of HTTP Methods but they are ' \
             'limited across endpoints. Please review our docs.'

    def __init__(self, **kwargs):

        super(MethodNotAllowed, self).__init__(**{
            'code': 'method_not_allowed',
            'detail': self.DETAIL,
            'status': falcon.HTTP_405,
            'title': 'HTTP Method not allowed',
        })


"""
    406 Not Acceptable
    ~~~~~~~~~~~~~~~~~~
"""


class RequestNotAcceptable(APIException):
    """ The request requires a supported Accept header

    This exception supports an array of supported Accept headers
    to be passed to the caller upon failure.
    """

    DETAIL = 'The URL requested (route) does not support the Accept ' \
             'header (mimetype) you provided. Supported mimetypes for ' \
             'the route are: {}. Also, a blank mimetype is equivalent to ' \
             '"*/*" & is acceptable. Please update your Accept header & ' \
             'retry your request.'

    def __init__(self, mimetypes, **kwargs):

        mimetypes = ', '.join('"{0}"'.format(m) for m in mimetypes)
        super(RequestNotAcceptable, self).__init__(**{
            'code': 'request_not_acceptable',
            'detail': self.DETAIL.format(mimetypes),
            'links': 'tools.ietf.org/html/rfc7231#section-5.3.2',
            'status': falcon.HTTP_406,
            'title': 'Unacceptable response preference',
        })


"""
    409 Conflict
    ~~~~~~~~~~~~
"""


class ResourceConflict(APIException):
    """ Generic resource conflict """

    DETAIL = 'Your request had a generic resource conflict.'

    def __init__(self, **kwargs):

        super(ResourceConflict, self).__init__(**{
            'code': 'resource_conflict',
            'detail': kwargs.get('detail', self.DETAIL),
            'status': falcon.HTTP_409,
            'title': 'Generic resource conflict',
        })


class ResourceTypeNotAllowed(APIException):
    """ The resource type is not allowed for the URL requested """

    DETAIL = 'You tried saving a resource to a route that doesn\'t ' \
             'match the type in your request payload. Please alter ' \
             'the route or the type in your payload & retry your request.'

    def __init__(self, **kwargs):

        super(ResourceTypeNotAllowed, self).__init__(**{
            'code': 'resource_type_not_allowed',
            'detail': self.DETAIL,
            'status': falcon.HTTP_409,
            'title': 'Resource type not allowed for this route',
        })


"""
    411 Length Required
    ~~~~~~~~~~~~~~~~~~~
"""


class ContentLengthRequired(APIException):
    """ The request lacks any indication of body length """

    DETAIL = 'Our API requires a Content-Length header according ' \
             'to the guidelines of RFC 7230 (value >= 0). Empty ' \
             'bodied requests should have a Content-Length of 0'

    def __init__(self, **kwargs):

        super(ContentLengthRequired, self).__init__(**{
            'code': 'length_required',
            'detail': self.DETAIL,
            'links': 'tools.ietf.org/html/rfc7230#section-3.3.2',
            'status': falcon.HTTP_411,
            'title': 'A Content-Length header is required',
        })


"""
    414 URI Too Long
    ~~~~~~~~~~~~~~~~
"""


class URITooLong(APIException):
    """ The request URI exceeds goldman.config.MAX_URI_LENGTH

    This exception requires a value of maximum URI length to
    be passed in for a more detailed response.
    """

    DETAIL = 'The URI in your request has exceeded the maximum ' \
             'length of %s. There isn\'t much you can do except ' \
             'contact us if you believe this is an error.'

    def __init__(self, max_length, **kwargs):

        super(URITooLong, self).__init__(**{
            'code': 'uri_too_long',
            'detail': self.DETAIL.format(max_length),
            'links': 'tools.ietf.org/html/rfc7231#section-6.5.12',
            'status': falcon.HTTP_414,
            'title': 'URI is too long',
        })


"""
    415 Unsupported Media Type
    ~~~~~~~~~~~~~~~~~~~~~~~~~~
"""


class ContentTypeUnsupported(APIException):
    """ The request requires a supported Content-Type header

    This exception supports an array of supported Content-Type's
    to be passed to the caller upon failure.
    """

    DETAIL = 'The URL requested (route) & the HTTP method used ' \
             'require a Content-Type header. Unfortunately, we ' \
             'couldn\'t find the header in your request or the one ' \
             'provided is unsupported. Supported content types for ' \
             'the route are: {}. Please update your Content-Type & ' \
             'retry your request.'

    def __init__(self, mimetypes, **kwargs):

        mimetypes = ', '.join('"{0}"'.format(m) for m in mimetypes)
        super(ContentTypeUnsupported, self).__init__(**{
            'code': 'content_type_required',
            'detail': self.DETAIL.format(mimetypes),
            'links': 'tools.ietf.org/html/rfc7231#section-6.5.13',
            'status': falcon.HTTP_415,
            'title': 'Content-type header missing & required or unsupported',
        })


"""
    417 Expectation Failed
    ~~~~~~~~~~~~~~~~~~~~~~
"""


class ExpectationUnmet(APIException):
    """ The request included an Expect header that is unsupported """

    DETAIL = 'Our API does not currently support any implementation ' \
             'of the Expect header. Please remove the header & retry ' \
             'your request per RFC 7231.'

    def __init__(self, **kwargs):

        super(ExpectationUnmet, self).__init__(**{
            'code': 'expectation_unmet',
            'detail': self.DETAIL,
            'links': 'tools.ietf.org/html/rfc7231#section-5.1.1',
            'status': falcon.HTTP_417,
            'title': 'Expectation cannot be met',
        })


"""
    429 Too Many Requests
    ~~~~~~~~~~~~~~~~~~~~~
"""


class TooManyRequests(APIException):
    """ The client has exceeded the allotment of requests """

    DETAIL = 'You have exceeded the request limit of {0} requests ' \
             'per {1} seconds. Please try again after {2} seconds ' \
             'have elapsed & let us know if this threshold is too ' \
             'conservative.'

    def __init__(self, **kwargs):

        # XXX FIX: falcon 4.0 has falcon.HTTP_429 support
        super(TooManyRequests, self).__init__(**{
            'code': 'too_many_requests',
            'detail': self.DETAIL,
            'links': 'tools.ietf.org/html/rfc6585#section-4',
            'status': '429 Too Many Requests',
            'title': 'You are being rate-limited',
        })


"""
    422 Unprocessable Entity
    ~~~~~~~~~~~~~~~~~~~~~~~~~~
"""


class ValidationFailure(APIException):
    """ The models validation routines failed when saving

    This exception supports detail & links overrides, however
    the source object containing a pointer to the attribute
    that caused the failure is required.
    """

    DETAIL = 'Generic validation error message'

    def __init__(self, pointer, **kwargs):
        super(ValidationFailure, self).__init__(**{
            'code': 'validation_failure',
            'detail': kwargs.get('detail', self.DETAIL),
            'links': kwargs.get('links'),
            'source': {'pointer': pointer},
            'status': '422 Unprocessable Entity',
            'title': 'Your attribute modification failed validations',
        })


"""
    500 Internal Server Error
    ~~~~~~~~~~~~~~~~~~~~~~~~~
"""


class GenericError(APIException):
    """ Something puked within our API & is probably 3rd party

    This exception supports detail & links overrides.
    """

    DETAIL = 'Generic internal server error'

    def __init__(self, **kwargs):
        super(GenericError, self).__init__(**{
            'code': 'generic_error',
            'detail': kwargs.get('detail', self.DETAIL),
            'links': kwargs.get('links'),
            'status': '500 Internal Server Error',
            'title': 'The request had an unexpected error',
        })


"""
    503 Service Unavailable
    ~~~~~~~~~~~~~~~~~~~~~~~
"""


class ServiceUnavailable(APIException):
    """ Something puked within our API, probably the database

    This exception supports detail & links overrides.
    """

    DETAIL = 'Generic service unavailable error'

    def __init__(self, **kwargs):
        super(ServiceUnavailable, self).__init__(**{
            'code': 'service_unavailable',
            'detail': kwargs.get('detail', self.DETAIL),
            'links': kwargs.get('links'),
            'status': '503 Service Unavailable',
            'title': 'Our server is currently unavailable',
        })


class DatabaseUnavailable(ServiceUnavailable):
    """ Something puked when accessing the database

    This is simply a convenience class
    """

    DETAIL = 'When handling your request we experienced an unexpected ' \
             'error when communicating with the database. This is likely ' \
             'transitory so please retry your request.'



"""
    resources.base
    ~~~~~~~~~~~~~~

    Base resource object.

    This should be sub-classed by other more mature resource
    handlers. It's utility is providing common tasks & error handling
    for all resources.
"""

import types


class Resource(object):
    """ Base resource class """

    DESERIALIZERS = []
    SERIALIZERS = []

    def __init__(self, disable=None):

        disable = disable or []
        rondrs = getattr(self, 'rondrs', [])

        if not self.DESERIALIZERS:
            raise NotImplementedError('resource DESERIALIZERS required')
        elif not self.SERIALIZERS:
            raise NotImplementedError('resource SERIALIZERS required')

        for rondr in rondrs:
            func = types.MethodType(rondr, self)
            name = rondr.func_name

            if name not in disable:
                setattr(self, name, func)

    @property
    def deserializer_mimetypes(self):
        """ Return a list of string mimetypes supported as deserializers """

        return [d.MIMETYPE for d in self.DESERIALIZERS]

    @property
    def serializer_mimetypes(self):
        """ Return a list of string mimetypes supported as serializers """

        return [s.MIMETYPE for s in self.SERIALIZERS]



"""
    types.datetime
    ~~~~~~~~~~~~~~

    schematics datetime type

    This should be used instead of the native schematics datetime
    type. It supports epoch serialization/deserialization in addition
    to the default behavior.
"""

from __future__ import absolute_import
from datetime import datetime as dt

from schematics.types import DateTimeType as SchematicsDateTimeType


class Type(SchematicsDateTimeType):
    """ Extended DateTimeType to support epoch timestamps """

    def to_primitive(self, value, context=None):
        """ Schematics serializer override

        If epoch_date is true then convert the `datetime.datetime`
        object into an epoch `int`.
        """

        if context and context.get('epoch_date'):
            epoch = dt(1970, 1, 1)
            value = (value - epoch).total_seconds()
            return int(value)
        elif context and context.get('datetime_date'):
            return value
        else:
            return super(Type, self).to_primitive(value, context)



"""
    utils.model_helpers
    ~~~~~~~~~~~~~~~~~~~

    Convient model helpers. That's it.
"""

import goldman


__all__ = ['rtype_to_model']


def rtype_to_model(rtype):
    """ Return a model class object given a string resource type

    :param rtype:
        string resource type
    :return:
        model class object
    :raise:
        ValueError
    """

    models = goldman.config.MODELS
    for model in models:
        if rtype.lower() == model.RTYPE.lower():
            return model
    raise ValueError('%s resource type not registered' % rtype)



"""
    stores
    ~~~~~~

    All stores supported by our API

    Be mindful of how they are named to avoid conflict with any
    official 3rd party packages.

    TIP: Unlike other modules in goldman we don't import all
         of the different stores since they need 3rd party
         libs. Load only if needed.
"""

from ..stores.base import Store as BaseStore


RESOURCES = [
    BaseStore,
]



"""
    serializers
    ~~~~~~~~~~~

    Location of different serializers supported by our API
"""

from ..serializers.base import Serializer as BaseSerializer
from ..serializers.comma_sep import Serializer as CsvSerializer
from ..serializers.json_7159 import Serializer as JsonSerializer
from ..serializers.jsonapi import Serializer as JsonApiSerializer


SERIALIZERS = [
    BaseSerializer,
    CsvSerializer,
    JsonSerializer,
    JsonApiSerializer,
]



"""
    types.to_one
    ~~~~~~~~~~~~

    schematics to_one type

    This should be used when defining a many-to-one or one-to-one
    relationship.
"""

import goldman
import goldman.validators as validators

from schematics.exceptions import ValidationError
from schematics.types import BaseType


class ToOne(object):
    """ ToOne object

    Evaluating whether or not to load the model from the
    store uses the is_loaded property rather than simply
    checking the model attribute. This is to needed when
    a load has been done but no model exists.
    """

    def __init__(self, rtype, field, rid=None):

        self.field = field
        self.rtype = rtype
        self.rid = rid

        self._is_loaded = False
        self.model = None

    def __eq__(self, other):

        try:
            return self.field == other.field and \
                self.rtype == other.rtype and \
                self.rid == other.rid
        except AttributeError:
            return False

    def __ne__(self, other):

        return not self == other

    def __repr__(self):

        name = self.__class__.__name__

        return '%s(\'%s\', \'%s\', rid=\'%s\')' % (name, self.rtype, self.field,
                                                   self.rid)

    def __str__(self):

        return str(self.rid)

    @property
    def is_loaded(self):
        """ Boolean indicating whether a load attempt has been made """

        return self._is_loaded

    def load(self):
        """ Return the model from the store """

        if self.rid and not self.is_loaded:
            store = goldman.sess.store
            self._is_loaded = True

            self.model = store.find(self.rtype, self.field, self.rid)

        return self.model


class Type(BaseType):
    """ Custom field for our ToOne relationships """

    MESSAGES = {
        'exists': 'item not found',
    }

    def __init__(self, field='rid', rtype=None, skip_exists=False,
                 typeness=int, **kwargs):

        self.field = field
        self.rtype = rtype

        self.skip_exists = skip_exists
        self.typeness = typeness

        super(Type, self).__init__(**kwargs)

    def _cast_rid(self, rid):
        """ Ensure the incoming resource ID is cast properly """

        if self.typeness is int:
            return int(rid)

    def to_native(self, value, context=None):
        """ Schematics deserializer override

        :return: ToOne instance
        """

        if isinstance(value, ToOne):
            return value

        value = self._cast_rid(value)
        return ToOne(self.rtype, self.field, rid=value)

    def to_primitive(self, value, context=None):
        """ Schematics serializer override

        :return: dict
        """

        if context and context.get('rel_ids'):
            return value.rid
        else:
            return {'rtype': value.rtype, 'rid': value.rid}

    def validate_to_one(self, value):
        """ Check if the to_one should exist & casts properly """

        if value.rid and self.typeness is int:
            validators.validate_int(value)

        if value.rid and not self.skip_exists:
            if not value.load():
                raise ValidationError(self.messages['exists'])
        return value



"""
    utils.url_helpers
    ~~~~~~~~~~~~~~~~~

    Interfaces that make constructing URL's a bit more easy.
    It's best not to scatter URL construction around in the code
    so it's hopefully easier to adjust if the API's routes are
    updated.

    If you are familiar with the Flask micro-web framework these
    interfaces will give you a url_for() type equivalent.
"""

import goldman


__all__ = ['rid_url', 'rtype_url']


def rtype_url(rtype):
    """ Return a string relative URL for a resource endpoint

    This represents the API endpoint for performing any
    supported interactions of a given model.

    :param rtype:
        string resource type name of the model
    :return:
        relative string URL
    """

    return '%s/%s' % (goldman.config.BASE_URL, rtype)


def rid_url(rtype, rid):
    """ Return a string relative URL for a unique resource id

    :param rtype:
        string resource type name of the model
    :param rid:
        string rid of the single model
    :return:
        relative string URL
    """

    return '%s/%s' % (rtype_url(rtype), rid)



"""
    resources
    ~~~~~~~~~

    Location of our different resources for Falcon.

    Resources are responsible for registering responders & tend
    to follow the same general workflows for our Mongo documents.
"""

from ..resources.base import Resource as BaseResource
from ..resources.json_7159 import Resource as JSONResource
from ..resources.model import Resource as ModelResource
from ..resources.models import Resource as ModelsResource
from ..resources.oauth_ropc import Resource as OAuthROPCResource
from ..resources.related import Resource as RelatedResource
from ..resources.s3_model import Resource as S3ModelResource
from ..resources.s3_model_image import Resource as S3ModelImageResource


RESOURCES = [
    BaseResource,
    JSONResource,
    ModelResource,
    ModelsResource,
    OAuthROPCResource,
    RelatedResource,
    S3ModelResource,
    S3ModelImageResource,
]



"""
    utils.str_helpers
    ~~~~~~~~~~~~~~~~~

    Convient string helpers. That's it.
"""

import hashlib
import uuid

from datetime import datetime as dt


def cmp_val_salt_hash(val, salt, str_hash):
    """ Given a string, salt, & hash validate the string

    The salt & val will be concatented as in gen_salt_and hash()
    & compared to the provided hash. This will only ever work
    with hashes derived from gen_salt_and_hash()

    :param val: clear-text string
    :param salt: string salt
    :param str_hash: existing hash to compare against
    :return: boolean
    """

    computed_hash = hashlib.sha256(val + salt).hexdigest()
    return computed_hash == str_hash


def gen_salt_and_hash(val=None):
    """ Generate a salt & hash

    If no string is provided then a random string will be
    used to hash & referred to as `val`.

    The salt will always be randomly generated & the hash
    will be a sha256 hex value of the `val` & the salt as
    a concatenated string. It follows the guidance here:

        crackstation.net/hashing-security.htm#properhashing

    :param val: str
    :return: tuple of strings (salt, hash)
    """

    if not val:
        val = random_str()

    str_salt = random_str()
    str_hash = hashlib.sha256(val + str_salt).hexdigest()
    return str_salt, str_hash


def naked(val):
    """ Given a string strip off all white space & quotes """

    return val.strip(' "\'\t')


def random_str():
    """ Return a random string """

    return str(uuid.uuid4())


def str_to_bool(val):
    """ Return a boolean if the string value represents one

    :param val: str
    :return: bool
    :raise: ValueError
    """

    if isinstance(val, bool):
        return val
    elif val.lower() == 'true':
        return True
    elif val.lower() == 'false':
        return False
    else:
        raise ValueError


def str_to_dt(val):
    """ Return a datetime object if the string value represents one

    Epoch integer or an ISO 8601 compatible string is
    supported.

    :param val: str
    :return: datetime
    :raise: ValueError
    """

    if isinstance(val, dt):
        return val

    try:
        if val.isdigit():
            return dt.utcfromtimestamp(float(val))
        else:
            return dt.strptime(val, '%Y-%m-%dT%H:%M:%S.%f')
    except (AttributeError, TypeError):
        raise ValueError



"""
    middleware.security
    ~~~~~~~~~~~~~~~~~~~

    High level HTTP standards type checks, enforcement of our
    own global rules, general security best practices, etc.

    This middleware should probably be hit as early as possible.
"""

import goldman

from goldman.exceptions import TLSRequired
from goldman.utils.error_helpers import abort


class Middleware(object):
    """ Falcon security middleware """

    # pylint: disable=unused-argument
    def process_request(self, req, resp):
        """ Process the request before routing it.

        We always enforce the use of SSL.
        """

        if goldman.config.TLS_REQUIRED and req.protocol != 'https':
            abort(TLSRequired)

    def process_response(self, req, resp, resource):
        """ Post-processing of the response (after routing).

        We attempt to follow many of the 'hardening reponse headers'
        best practices. These include things like:

            HPKP: HTTP Public Key Pinning
            HSTS: HTTP Srict Transport Security
            X-Content-Type-Options:
                prevents mime-sniffing so user generated content
                isn't auto determined as something it shouldn't be.
            XSS Protection: built in reflective XSS protection
        """

        hpkp = goldman.config.HPKP

        if hpkp:
            resp.set_header(
                'Public-Key-Pins',
                '{}; includeSubdomains; max-age=31536000'.format(hpkp),
            )

        resp.set_header(
            'Strict-Transport-Security',
            'max-age=31536000; includeSubDomains',
        )

        resp.set_header(
            'X-Content-Type-Options',
            'nosniff',
        )

        resp.set_header(
            'X-Xss-Protection',
            '1; mode=block',
        )



"""
    middleware.rate_limit
    ~~~~~~~~~~~~~~~~~~~~~

    Basic rate limiter for the API to avoid overloading the
    API by friend or asshole. It follows the guidance of
    RFC 6585 section 4 by leveraging HTTP 429 status codes &
    RFC 7231 section 7.1.3 Retry-After headers.

    The middleware looks for goldman.config constants by
    the name of:

        RATE_LIMIT_COUNT - the number of requests allowed
                           within a given RATE_LIMIT_DURATION

        RATE_LIMIT_DURATION - number of seconds long the window
                              of RATE_LIMIT_COUNT applies

    This middleware should probably be the very first
    middleware hit to avoid as much processing as possible.
"""

import goldman
import goldman.exceptions as exceptions

from cachetools import TTLCache
from goldman.utils.error_helpers import abort


class Middleware(object):
    """ Falcon rate limiting middleware """

    def __init__(self):

        self.count = goldman.config.RATE_LIMIT_COUNT
        self.duration = goldman.config.RATE_LIMIT_DURATION

        self.cache = TTLCache(maxsize=self.count, ttl=self.duration)

    @property
    def _error_headers(self):
        """ Return a dict of headers in every auth failure """

        return {
            'Retry-After': self.duration,
            'X-RateLimit-Limit': self.count,
            'X-RateLimit-Remaining': 0,
        }

    # pylint: disable=unused-argument
    def process_request(self, req, resp):
        """ Process the request before routing it. """

        key = req.env['REMOTE_PORT'] + req.env['REMOTE_ADDR']
        val = self.cache.get(key, 0)

        if val == self.count:
            abort(exceptions.TooManyRequests(headers=self._error_headers))
        else:
            self.cache[key] = val + 1



"""
    middleware.bearer_token
    ~~~~~~~~~~~~~~~~~~~~~~~

    OAuth 2.0 Authorization Framework: Bearer Token Usage
    implementation as documented in RFC 6750.

        Specifically, the middleware supports ONLY section
        2.1 of RFC 6750; an Authorization header in the
        request containing an unencoded token. It will
        come after the scheme is declared which is a string
        of "Bearer".

    This middleware requires a callable to be passed in as
    the `auth_token` property which will be given the token.

    The callable should return a login model of the successfully
    authenticated user or raise an `AuthRejected` exception
    causing the request to be aborted IF the middleware's
    `optional` property is set to False (default).

    The model will be assigned to the `goldman.sess.login`
    propery if authentication succeeds.

    NOTE: As documented in RFC 6750 on certain errors the
          the registered error code & description is passed
          in the WWW-Authenticate header. BUT ONLY SOMETIMES.

          The spec makes it clear that a response SHOULD
          NOT include the error info in the header if
          either no Authorization header is present or the
          wrong "scheme" is used.

          This ambiguity sucks for our clients & since the
          spec gives no guidance on a payload containing
          errors we ALWAYS include one in addition to
          following the section 3.1 nonsense.

    WARN: auth will be completely bypassed if the token_endpoint
          or revoke_endpoint are being accessed.

          This is so they can complete when a bearer token
          is not known or passed in differently.. don't
          forget the OAuth access_token resource!
"""

import goldman
import goldman.signals as signals

from goldman.exceptions import (
    AuthRejected,
    AuthRequired,
    InvalidAuthSyntax,
)
from goldman.utils.error_helpers import abort
from goldman.utils.str_helpers import naked


class Middleware(object):
    """ Ensure RFC compliance & authenticate the token. """

    def __init__(self, auth_token, optional=False, revoke_endpoint='/revoke',
                 token_endpoint='/token'):

        self.auth_token = auth_token
        self.optional = optional
        self.revoke_endpoint = revoke_endpoint
        self.token_endpoint = token_endpoint

    @property
    def _error_headers(self):
        """ Stock dict of headers when an error occurs """

        return {
            'Cache-Control': 'no-store',
            'Pragma': 'no-cache',
            'WWW-Authenticate': self._realm,
        }

    @property
    def _realm(self):
        """ Return a string representation of the authentication realm """

        return 'Bearer realm="%s"' % goldman.config.AUTH_REALM

    def _get_invalid_token_headers(self, desc):
        """ Return our default error headers with section 3.1 headers

        Specifically, the "invalid_token" error code in section 3.1
        """

        headers = self._error_headers
        headers['WWW-Authenticate'] += ', error="invalid_token", ' \
                                       'error_description="%s"' % desc

        return headers

    def _validate_auth_scheme(self, req):
        """ Check if the request has auth & the proper scheme

        Remember NOT to include the error related info in
        the WWW-Authenticate header for these conditions.

        :raise:
            AuthRequired
        """

        if not req.auth:
            raise AuthRequired(**{
                'detail': 'You must first login to access the requested '
                          'resource(s). Please retry your request using '
                          'OAuth 2.0 Bearer Token Authentication as '
                          'documented in RFC 6750. If you do not have an '
                          'access_token then request one at the token '
                          'endpdoint of: %s' % self.token_endpoint,
                'headers': self._error_headers,
                'links': 'tools.ietf.org/html/rfc6750#section-2.1',
            })
        elif req.auth_scheme != 'bearer':
            raise AuthRequired(**{
                'detail': 'Your Authorization header is using an unsupported '
                          'authentication scheme. Please modify your scheme '
                          'to be a string of: "Bearer".',
                'headers': self._error_headers,
                'links': 'tools.ietf.org/html/rfc6750#section-2.1',
            })

    def _get_token(self, req):
        """ Get the token from the Authorization header

        If the header is actually malformed where Bearer Auth was
        indicated by the request then an InvalidAuthSyntax exception
        is raised. Otherwise an AuthRequired exception since it's
        unclear in this scenario if the requestor was even aware
        Authentication was required & if so which "scheme".

        Calls _validate_auth_scheme first & bubbles up it's
        exceptions.

        :return:
            string token
        :raise:
            AuthRequired, InvalidAuthSyntax
        """

        self._validate_auth_scheme(req)

        try:
            return naked(req.auth.split(' ')[1])
        except IndexError:
            desc = 'You are using the Bearer Authentication scheme as ' \
                   'required to login but your Authorization header is ' \
                   'completely missing the access_token.'

            raise InvalidAuthSyntax(**{
                'detail': desc,
                'headers': self._get_invalid_token_headers(desc),
                'links': 'tools.ietf.org/html/rfc6750#section-2.1',
            })

    def process_request(self, req, resp):  # pylint: disable=unused-argument
        """ Process the request before routing it. """

        if req.path in (self.revoke_endpoint, self.token_endpoint):
            return

        signals.pre_authenticate.send()

        try:
            token = self._get_token(req)
            goldman.sess.login = self.auth_token(token)
            signals.post_authenticate.send()
        except (AuthRequired, InvalidAuthSyntax) as exc:
            if not self.optional:
                abort(exc)
        except AuthRejected as exc:
            if not self.optional:
                exc.headers = self._get_invalid_token_headers(exc.detail)
                abort(exc)



"""
    serializers.jsonapi_error
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Serializer that is compliant with the JSON API specification
    for error handling as defined here:

        jsonapi.org/format/#errors
"""

import copy
import falcon
import goldman
import json

from falcon.http_status import HTTPStatus


class Serializer(HTTPStatus):
    """ JSON API compliant serializer for exceptions

    ERROR_OBJECT_FIELDS contains all of the JSON API compliant
    fields that are allowed. They are listed in the same exact
    order as they are on the specifications page.

    Fields present but not compatible with the specification are
    silently removed.

    :param errors:
        APIException object or list of them
    """

    ERROR_OBJECT_FIELDS = [
        'id',
        'links',
        'status',
        'code',
        'title',
        'detail',
        'source',
        'meta',
    ]

    def __init__(self, errors):

        if not isinstance(errors, list):
            errors = [errors]

        self.errors = [e().to_dict() for e in errors]
        super(Serializer, self).__init__(
            self.get_status(),
            body=self.get_body(),
            headers=self.get_headers()
        )

    def get_body(self):
        """ Return a HTTPStatus compliant body attribute

        Be sure to purge any unallowed properties from the object.

        TIP: At the risk of being a bit slow we copy the errors
             instead of mutating them since they may have key/vals
             like headers that are useful elsewhere.
        """

        body = copy.deepcopy(self.errors)

        for error in body:
            for key in error.keys():
                if key not in self.ERROR_OBJECT_FIELDS:
                    del error[key]
        return json.dumps({'errors': body})

    def get_headers(self):
        """ Return a HTTPStatus compliant headers attribute

        FIX: duplicate headers will collide terribly!
        """

        headers = {'Content-Type': goldman.JSONAPI_MIMETYPE}

        for error in self.errors:
            if 'headers' in error:
                headers.update(error['headers'])
        return headers

    def get_status(self):
        """ Return a HTTPStatus compliant status attribute

        Per the JSON API spec errors could have different status
        codes & a generic one should be chosen in these conditions
        for the actual HTTP response code.
        """

        codes = [error['status'] for error in self.errors]
        same = all(code == codes[0] for code in codes)

        if not same and codes[0].startswith('4'):
            return falcon.HTTP_400
        elif not same and codes[0].startswith('5'):
            return falcon.HTTP_500
        else:
            return codes[0]



"""
    resources.models
    ~~~~~~~~~~~~~~~~

    Multiple models resource object with responders.
"""

import falcon
import goldman
import goldman.signals as signals

from ..resources.base import Resource as BaseResource
from goldman.utils.responder_helpers import (
    from_rest,
    to_rest_model,
    to_rest_models,
)


def on_get(resc, req, resp):
    """ Get the models identified by query parameters

    We return an empty list if no models are found.
    """

    signals.responder_pre_any.send(resc.model)
    signals.responder_pre_search.send(resc.model)

    models = goldman.sess.store.search(resc.rtype, **{
        'filters': req.filters,
        'pages': req.pages,
        'sorts': req.sorts,
    })

    props = to_rest_models(models, includes=req.includes)
    resp.serialize(props)

    signals.responder_post_any.send(resc.model)
    signals.responder_post_search.send(resc.model)


def on_post(resc, req, resp):
    """ Deserialize the payload & create the new single item """

    signals.responder_pre_any.send(resc.model)
    signals.responder_pre_create.send(resc.model)

    props = req.deserialize()
    model = resc.model()

    from_rest(model, props)
    goldman.sess.store.create(model)

    props = to_rest_model(model, includes=req.includes)
    resp.last_modified = model.updated
    resp.location = '%s/%s' % (req.path, model.rid_value)
    resp.status = falcon.HTTP_201
    resp.serialize(props)

    signals.responder_post_any.send(resc.model)
    signals.responder_post_create.send(resc.model)


class Resource(BaseResource):
    """ Multiple items resource & responders """

    DESERIALIZERS = [
        goldman.JsonApiDeserializer,
    ]

    SERIALIZERS = [
        goldman.CsvSerializer,
        goldman.JsonApiSerializer,
    ]

    def __init__(self, model, disable=None):

        self.model = model
        self.rondrs = [on_get, on_post]
        self.rtype = model.RTYPE

        super(Resource, self).__init__(disable)



"""
    utils.decorators
    ~~~~~~~~~~~~~~~~

    Our custom python decorators.
"""


# pylint: disable=invalid-name
class classproperty(object):  # NOQA
    """ @classmethod meets @property

    TIP: This is only a getter! Not a setter!

    Inspired by stackoverflow:
        /questions/5189699/how-can-i-make-a-class-property-in-python
    """
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)



"""
    models.default_schema
    ~~~~~~~~~~~~~~~~~~~~~

    Using the default_schema assumes a few things:

        * the API uses authentication
        * each model has a creator field defaulting to the
          logged in user that created the model
        * each model has timestamp fields for create / update

    It will also use the goldman BaseModel as is required on
    all models if using a goldman store.
"""

import goldman
import goldman.signals as signals

from ..models.base import Model as BaseModel
from datetime import datetime as dt
from goldman.types import DateTimeType, ToOneType
from schematics.types import IntType


class Model(BaseModel):
    """ Model with some common schematics types """

    creator = ToOneType(
        from_rest=False,
        rtype='logins',
        skip_exists=True,
    )

    created = DateTimeType(from_rest=False)
    updated = DateTimeType(from_rest=False)

    rid = IntType(
        from_rest=False,
        rid=True,
    )


# pylint: disable=unused-argument
def pre_create(sender, model):
    """ Callback before creating any new model

    Identify the creator of the new model & set the
    created timestamp to now.
    """

    model.created = dt.utcnow()
    model.creator = goldman.sess.login


def pre_save(sender, model):
    """ Callback before saving any model

    Update the updated timestamp to now.
    """

    model.updated = dt.utcnow()


signals.pre_create.connect(pre_create)
signals.pre_save.connect(pre_save)



"""
    models
    ~~~~~~

    Module containing all of our models that are typically
    accessed in a CRUD like manner.
"""

from ..models.base import Model as BaseModel
from ..models.default_schema import Model as DefaultSchemaModel
from ..models.login import Model as LoginModel


MODELS = [
    BaseModel,
    DefaultSchemaModel,
    LoginModel,
]



"""
    deserializers.form_urlencoded
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Deserializer that simply acknowledges the need for form
    uploads according to:

        www.w3.org/TR/1999/REC-html401-19991224/interact/forms.html#h-17.13.4

    Falcon already takes form-urlencoded params & makes them
    available via request.get_param() so there is no need to
    do anything except register the mimetype with the deserializer.
"""

import goldman

from ..deserializers.base import Deserializer as BaseDeserializer


class Deserializer(BaseDeserializer):
    """ Form URL encoded compliant deserializer """

    MIMETYPE = goldman.FORMURL_MIMETYPE



"""
    postgres.store
    ~~~~~~~~~~~~~~

    Interface with helper routines for persisting, finding,
    etc from a postgresql database.
"""

import goldman
import goldman.exceptions as exceptions
import goldman.signals as signals

from ..base import Store as BaseStore
from ..postgres.connect import Connect
from goldman.queryparams.filter import FilterOr, FilterRel
from goldman.queryparams.sort import Sortable
from goldman.utils.error_helpers import abort
from goldman.utils.model_helpers import rtype_to_model


CONNECT = Connect()


ERRORS_TABLE = {
    '42883': 'One or more of the query filters had an unexpected value '
             'that could not be processed. This is a rare scenario & '
             'it is unclear which field has the issue. Please review '
             'your filters for any irregularities.',
}


FILTER_TABLE = {
    'eq': '=',
    'gt': '>',
    'lt': '<',
    'ne': '<>',
    'gte': '>=',
    'lte': '<=',

    'contains': 'LIKE',
    'icontains': 'ILIKE',
    'endswith': 'LIKE',
    'startswith': 'LIKE',

    'in': 'IN',
    'nin': 'NOT IN',

    'near': '<=',

    'after': '>',
    'before': '<',
}


def handle_exc(exc):
    """ Given a database exception determine how to fail

    Attempt to lookup a known error & abort on a meaningful
    error. Otherwise issue a generic DatabaseUnavailable exception.

    :param exc: psycopg2 exception
    """

    err = ERRORS_TABLE.get(exc.pgcode)

    if err:
        abort(exceptions.InvalidQueryParams(**{
            'detail': err,
            'parameter': 'filter',
        }))

    abort(exceptions.DatabaseUnavailable)


class Store(BaseStore):
    """ PostgreSQL database store """

    def __init__(self):

        self.conn = CONNECT.connect()

        super(Store, self).__init__()

    @staticmethod
    def dirty_cols(model):
        """ Get the models dirty columns in a friendly SQL format

        This will be a string of comma separated field names
        prefixed by the models resource type. If no fields are
        dirty then return None

        :return: str or None
        """

        cols = ', '.join(model.dirty_fields)

        return cols or None

    @staticmethod
    def dirty_vals(model):
        """ Get the models dirty values in a friendly SQL format

        This will be a string of comma separated field
        names in a format for psycopg to substitute with
        parameterized inputs.

        An example, if the `rid` & `created` fields are
        dirty then they'll be converted into:

            '%(rid)s, %(created)s'

        :return: str or None
        """

        vals = []

        for field in model.dirty_fields:
            vals.append('%({0})s'.format(field))

        vals = ', '.join(vals)

        return vals or None

    @staticmethod
    def field_cols(model):
        """ Get the models columns in a friendly SQL format

        This will be a string of comma separated field
        names prefixed by the models resource type.

        TIP: to_manys are not located on the table in Postgres
             & are instead application references, so any reference
             to there column names should be pruned!

        :return: str
        """

        to_many = model.to_many
        cols = [f for f in model.all_fields if f not in to_many]
        cols = ', '.join(cols)

        return cols or None

    @staticmethod
    def filters_query(filters):
        """ Turn the tuple of filters into SQL WHERE statements

        The key (column name) & operator have already been vetted
        so they can be trusted but the value could still be evil
        so it MUST be a parameterized input!

        That is done by creating a param dict where they key name
        & val look like:

            '{}_{}'.format(key, oper): val

        The key is constructed the way it is to ensure uniqueness,
        if we just used the key name then it could get clobbered.

        Ultimately the WHERE statement will look something like:

            age >= {age_gte}

        where age_gte is the key name in the param dict with a
        value of the evil user input. In the end, a string
        statement & dict param are returned as a tuple if any
        filters were provided otherwise None.

        :return: tuple (string, dict)
        """

        def _cast_val(filtr):
            """ Perform any needed casting on the filter value

            This could be tasks like including '%' signs at
            certain anchor points based on the filter or
            even wrapping it in certain functions.
            """

            val = filtr.val

            if filtr.oper in ('contains', 'icontains'):
                val = '%' + filtr.val + '%'
            elif filtr.oper == 'endswith':
                val = '%' + filtr.val
            elif filtr.oper == 'startswith':
                val = filtr.val + '%'

            return val

        def _filter(filtr):
            """ Process each individual Filter object """

            oper = FILTER_TABLE[filtr.oper]
            prop = '{field}_{oper}'.format(
                field=filtr.field.replace('.', '_'),
                oper=filtr.oper,
            )

            if isinstance(filtr, FilterRel):
                stmt = _filter_rel(filtr, oper, prop)
            else:
                stmt = '{field} {oper} %({prop})s'.format(
                    field=filtr.field,
                    oper=oper,
                    prop=prop,
                )

            return stmt, {prop: _cast_val(filtr)}

        def _filter_or(filters):
            """ Given a FilterOr object return a SQL query """

            param = {}
            stmts = []

            for filtr in filters:
                vals = _filter(filtr)

                param.update(vals[1])
                stmts.append(vals[0])

            stmt = ' OR '.join(stmts)
            stmt = '({})'.format(stmt)

            return stmt, param

        def _filter_rel(rel, oper, prop):
            """ Given a FilterRel object return a SQL sub query """

            stmt = """
                   {field} IN (SELECT {foreign_field} FROM {foreign_rtype}
                               WHERE {foreign_filter} {oper} %({prop})s)
                   """

            return stmt.format(
                field=rel.local_field,
                foreign_field=rel.foreign_field,
                foreign_filter=rel.foreign_filter,
                foreign_rtype=rel.foreign_rtype,
                oper=oper,
                prop=prop,
            )

        param = {}
        stmts = []

        for filtr in filters:
            if isinstance(filtr, FilterOr):
                vals = _filter_or(filtr)
            else:
                vals = _filter(filtr)

            param.update(vals[1])
            stmts.append(vals[0])

        if stmts:
            stmt = ' AND '.join(stmts)
            stmt = ' WHERE ' + stmt

        return stmt, param

    @staticmethod
    def pages_query(pages):
        """ Turn the tuple of pages into a SQL LIMIT/OFFSET query """

        try:
            return ' OFFSET {0} LIMIT {1}'.format(pages.offset, pages.limit)
        except AttributeError:
            return ''

    @staticmethod
    def sorts_query(sortables):
        """ Turn the Sortables into a SQL ORDER BY query """

        stmts = []

        for sortable in sortables:
            if sortable.desc:
                stmts.append('{} DESC'.format(sortable.field))
            else:
                stmts.append('{} ASC'.format(sortable.field))

        return ' ORDER BY {}'.format(', '.join(stmts))

    @staticmethod
    def to_pg(model):
        """ Invoke the models to_primitive properly for pg """

        return model.to_primitive(context={
            'datetime_date': True,
            'rel_ids': True,
        })

    def create(self, model):
        """ Given a model object instance create it """

        signals.pre_create.send(model.__class__, model=model)
        signals.pre_save.send(model.__class__, model=model)

        param = self.to_pg(model)
        query = """
                INSERT INTO {table} ({dirty_cols})
                VALUES ({dirty_vals})
                RETURNING {cols};
                """

        query = query.format(
            cols=self.field_cols(model),
            dirty_cols=self.dirty_cols(model),
            dirty_vals=self.dirty_vals(model),
            table=model.rtype,
        )

        result = self.query(query, param=param)

        signals.post_create.send(model.__class__, model=model)
        signals.post_save.send(model.__class__, model=model)

        return model.merge(result[0], clean=True)

    def delete(self, model):
        """ Given a model object instance delete it """

        signals.pre_delete.send(model.__class__, model=model)

        param = {'rid_value': self.to_pg(model)[model.rid_field]}
        query = """
                DELETE FROM {table}
                WHERE {rid_field} = %(rid_value)s
                RETURNING {cols};
                """

        query = query.format(
            cols=self.field_cols(model),
            rid_field=model.rid_field,
            table=model.rtype,
        )

        result = self.query(query, param=param)

        signals.post_delete.send(model.__class__, model=model)

        return result

    def find(self, rtype, key, val):
        """ Given a resource type & a single key/val find the model

        A single instantiated model object found by the key/val
        will be returned if a match is found in the database.
        If no record can be found then return None.

        WARN: This isn't for complex queries! Use search() instead.

        :return: model or None
        """

        model = rtype_to_model(rtype)
        param = {'key': key, 'val': val}
        query = """
                SELECT {cols} FROM {table}
                WHERE {key} = %(val)s;
                """

        query = query.format(
            cols=self.field_cols(model),
            key=key,
            table=rtype,
        )

        signals.pre_find.send(model.__class__, model=model)

        result = self.query(query, param=param)
        if result:
            result = model(result[0])
            signals.post_find.send(model.__class__, model=result)

        return result or None

    def query(self, query, param=None):
        """ Perform a SQL based query

        This will abort on a failure to communicate with
        the database.

        :query: string query
        :params: parameters for the query
        :return: RecordList from psycopg2
        """

        with self.conn.cursor() as curs:
            print 'XXX QUERY', curs.mogrify(query, param)
            try:
                curs.execute(query, param)
            except BaseException as exc:
                msg = 'query: {}, param: {}, exc: {}'.format(query, param, exc)

                if hasattr(exc, 'pgcode'):
                    msg = '{}, exc code: {}'.format(msg, exc.pgcode)

                print msg
                handle_exc(exc)

            result = curs.fetchall()

        return result

    def search(self, rtype, **kwargs):
        """ Search for the model by assorted criteria

        Quite a bit needs to happen for a search processing!

        The breakdown is we need to apply the query parameters
        where applicable. These include pagination, sorting,
        & filtering.

        Additionally, models can declare there own static
        filters or static query that should be applied. For
        static filters a model can opt-in with a `search_filters`
        property & for a static query to append a `search_query`
        property must be present.
        """

        model = rtype_to_model(rtype)
        param = {}
        pages = self.pages_query(kwargs.get('pages'))
        sorts = self.sorts_query(kwargs.get(
            'sorts', [Sortable(goldman.config.SORT)]
        ))

        query = """
                SELECT {cols}, count(*) OVER() as _count
                FROM {table}
                """
        query = query.format(
            cols=self.field_cols(model),
            table=rtype,
        )

        filters = kwargs.get('filters', [])
        filters += getattr(model, 'search_filters', []) or []

        if filters:
            where, param = self.filters_query(filters)
            query += where

        model_query = getattr(model, 'search_query', '') or ''
        if filters and model_query:
            model_query = ' AND ' + model_query
        elif model_query:
            model_query = ' WHERE ' + model_query

        query += model_query
        query += sorts
        query += pages

        signals.pre_search.send(model.__class__, model=model)

        result = self.query(query, param=param)
        models = [model(res) for res in result]

        if models:
            signals.post_search.send(model.__class__, models=result)

        pages = kwargs.get('pages')
        if pages and result:
            pages.total = result[0]['_count']

        return models

    def update(self, model):
        """ Given a model object instance update it """

        signals.pre_update.send(model.__class__, model=model)
        signals.pre_save.send(model.__class__, model=model)

        param = self.to_pg(model)
        param['rid_value'] = param[model.rid_field]

        query = """
                UPDATE {table}
                SET ({dirty_cols}) = ({dirty_vals})
                WHERE {rid_field} = %(rid_value)s
                RETURNING {cols};
                """

        query = query.format(
            cols=self.field_cols(model),
            dirty_cols=self.dirty_cols(model),
            dirty_vals=self.dirty_vals(model),
            rid_field=model.rid_field,
            table=model.rtype,
        )

        result = self.query(query, param=param)

        signals.post_update.send(model.__class__, model=model)
        signals.post_save.send(model.__class__, model=model)

        return model.merge(result[0], clean=True)



"""
    middleware.http_specs
    ~~~~~~~~~~~~~~~~~~~~~

    This middleware is used to implement certain HTTP/1.1
    semantics with the following objectives:

        * API consistent error messages with meaningful
          error messages for remediation

        * inform clients of certain behavior they are
          exhibiting not in-line with HTTP/1.1 MUST's &
          SHOULD's

        * inform clients of certain optional HTTP/1.1
          guidelines that are unsupported by our API

        * certain security type safe guards


    The RFC's that influence this middleware are the 72xx
    guidelines. 7230, 7231, etc.

    This middleware should probably be hit as early as possible
    but after any security related checks are performed.
"""

import falcon
import goldman
import goldman.exceptions as exceptions

from goldman.utils.error_helpers import abort


class Middleware(object):
    """ Falcon HTTP 1.1 middleware """

    # pylint: disable=unused-argument
    def process_request(self, req, resp):
        """ Process the request before routing it.

        A summary of the rational for the different behavior
        is shown below.

        Expect Header
        ~~~~~~~~~~~~~

        Our API does not currently support any implementation
        of the Expect header. This mostly impacts clients that
        use the `100-continue` expectation. Per the spec a
        response code of HTTP 417 should force the client to
        retry without the header.

            tools.ietf.org/html/rfc7231#section-5.1.1

        URI Too Long
        ~~~~~~~~~~~~

        99.99% of the time this is a protection mechanism
        against assholes doing asshole'ish stuff. 0.01% of
        the time it's a bug.

        Message Body Length
        ~~~~~~~~~~~~~~~~~~~

        RFC 7230 section 3.3.3 has everything you really need
        to know. Per the spec a Content-Length header & a
        Transfer-Encoding header are mutaully exclusive & if
        neither are provided then return a 411 indicating a
        Content-Length header is required.

        NOTE: falcon already raises an exception for invalid
              Content-Length values so we don't need to type
              check it.
        """

        max_length = goldman.config.MAX_URI_LENGTH

        if req.expect:
            abort(exceptions.ExpectationUnmet)
        elif len(req.uri) > max_length:
            abort(exceptions.URITooLong(max_length))
        elif req.content_required:
            # content_length could be zero
            has_content_length = req.content_length is not None
            transfer_encoding = req.get_header('transfer-encoding')

            if has_content_length and transfer_encoding:
                abort(exceptions.InvalidRequestHeader(**{
                    'detail': 'Your request has both Transfer-Encoding & '
                              'Content-Length headers which is a violation '
                              'of RFC 7230. This is likely a bug in your '
                              'client & your request cannot proceed as is.',
                    'links': 'tools.ietf.org/html/rfc7230#section-3.3.3',
                }))
            elif transfer_encoding:
                pass
            elif not has_content_length:
                abort(exceptions.ContentLengthRequired)

    def process_response(self, req, resp, resource):
        """ Post-processing of the response (after routing).

        Some fundamental errors can't be intercepted any other
        way in Falcon. These include 404 when the route isn't found
        & 405 when the method is bunk. Falcon will try to send its
        own errors.

        In these cases, intercept them & replace them with our
        JSON API compliant version.

        TIP: If no route could be determined then the resource
             will be None.
        """

        if not resource and resp.status == falcon.HTTP_404:
            abort(exceptions.RouteNotFound)
        elif resp.status == falcon.HTTP_405:
            abort(exceptions.MethodNotAllowed)



"""
    queryparams.sort
    ~~~~~~~~~~~~~~~~

    Sort resources according to one or more criteria.

    The query parameter itself follow the JSON API convention
    for sorting fields:

        jsonapi.org/format/#fetching-sorting
"""

import goldman

from goldman.exceptions import InvalidQueryParams


LINK = 'jsonapi.org/format/#fetching-sorting'
PARAM = 'sort'


class Sortable(object):
    """ Sortable object

    An instance of this object can be used to express a
    sorting critera for later use by the store or whatever.

    If the sorting field name begins with a '-' then it will sort
    in descending order, otherwise ascending.

    :param field:
        String field name to sort on. Required.
    """

    def __init__(self, field):
        """ Preserve the original field for easy compares & output """

        self.field = field.lstrip('-')
        self.raw_field = field

    @property
    def asc(self):
        """ Boolean of ascending sort preference """

        return not self.raw_field.startswith('-')

    @property
    def desc(self):
        """ Boolean of descending sort preference """

        return self.raw_field.startswith('-')

    def __eq__(self, other):
        """ Compare other Sortable objects """

        try:
            return self.raw_field == other.raw_field
        except AttributeError:
            return False

    def __repr__(self):
        """ Display the original sort with descending character """

        name = self.__class__.__name__

        return '%s(\'%s\')' % (name, self.raw_field)

    def __str__(self):
        """ Display the original sort with descending character """

        return self.raw_field


def _validate_field(param, fields):
    """ Ensure the sortable field exists on the model """

    if param.field not in fields:
        raise InvalidQueryParams(**{
            'detail': 'The sort query param value of "%s" is '
                      'invalid. That field does not exist on the '
                      'resource being requested.' % param.raw_field,
            'links': LINK,
            'parameter': PARAM,
        })


def _validate_no_rels(param, rels):
    """ Ensure the sortable field is not on a relationship """

    if param.field in rels:
        raise InvalidQueryParams(**{
            'detail': 'The sort query param value of "%s" is not '
                      'supported. Sorting on relationships is not '
                      'currently supported' % param.raw_field,
            'links': LINK,
            'parameter': PARAM,
        })


def init(req, model):
    """ Determine the sorting preference by query parameter

    Return an array of Sortable objects.
    """

    rels = model.relationships
    fields = model.all_fields

    params = req.get_param_as_list('sort') or [goldman.config.SORT]
    params = [Sortable(param.lower()) for param in params]

    for param in params:
        _validate_no_rels(param, rels)
        _validate_field(param, fields)

    return params



"""
    models.base
    ~~~~~~~~~~~

    Our schematics sub-classed model required by all models
    that use any of our goldman native stores.

    This is required when creating any models that use a
    goldman store. The API on this model is expected to be
    available on every model being pushed through or pulled
    from a store.
"""

from goldman.exceptions import ValidationFailure
from goldman.types import ResourceType, ToManyType, ToOneType
from goldman.utils.decorators import classproperty
from goldman.utils.error_helpers import abort
from schematics.exceptions import ConversionError, ModelValidationError
from schematics.models import Model as _SchematicsModel
from schematics.types import EmailType


class Model(_SchematicsModel):
    """ Our schematics sub-classed model """

    def __init__(self, data=None, **kwargs):

        super(Model, self).__init__(data, strict=False, **kwargs)

        try:
            self._original = self.to_native()
        except BaseException:
            self._original = {}

    def __setattr__(self, name, value):
        """ Help auto-cast certain types since schematics doesn't

        We may want to auto-cast more field types on assignment
        but for now ToOneType's will properly auto-cast a
        model assignment.

        This supports assignment of a model directly to a
        ToOneType field where the `rid_value` property of the
        model will be extracted just as the `to_native` method
        of the ToOneType expects.
        """

        print 'XXX IS THIS NEEDED?'
        if name in self.to_one and hasattr(value, 'rid_value'):
            to_one = getattr(self, '_fields')[name]
            value = to_one.to_native(value.rid_value)

        super(Model, self).__setattr__(name, value)

    @classproperty
    def all_fields(cls):  # NOQA
        """ Return a list of all the fields """

        fields = getattr(cls, '_fields')

        return fields.keys()

    @classproperty
    def relationships(cls):  # NOQA
        """ Return a list of all the fields that are relationships """

        return cls.to_many + cls.to_one

    @classproperty
    def rid_field(cls):  # NOQA
        """ Return the resource id field str name """

        return cls.get_fields_by_prop('rid', True)[0]

    @classproperty
    def rtype_field(cls):  # NOQA
        """ Return the resource type field str name """

        return cls.get_fields_by_class(ResourceType)[0]

    @classproperty
    def to_lower(cls):  # NOQA
        """ Return a list of all the fields that should be lowercased

        This is done on fields with `lower=True`.
        """

        email = cls.get_fields_by_class(EmailType)
        lower = cls.get_fields_by_prop('lower', True) + email

        return list(set(email + lower))

    @classproperty
    def to_many(cls):  # NOQA
        """ Return a list of all the ToMany field types """

        return cls.get_fields_by_class(ToManyType)

    @classproperty
    def to_one(cls):  # NOQA
        """ Return a list of all the ToOne field types """

        return cls.get_fields_by_class(ToOneType)

    @classmethod
    def get_fields_by_class(cls, field_class):
        """ Return a list of field names matching a field class

        :param field_class: field class object
        :return: list
        """

        ret = []

        for key, val in getattr(cls, '_fields').items():
            if isinstance(val, field_class):
                ret.append(key)
        return ret

    @classmethod
    def get_fields_by_prop(cls, prop_key, prop_val):
        """ Return a list of field names matching a prop key/val

        :param prop_key: key name
        :param prop_val: value
        :return: list
        """

        ret = []

        for key, val in cls.get_fields_with_prop(prop_key):
            if val == prop_val:
                ret.append(key)
        return ret

    @classmethod
    def get_fields_with_prop(cls, prop_key):
        """ Return a list of fields with a prop key defined

        Each list item will be a tuple of field name containing
        the prop key & the value of that prop key.

        :param prop_key: key name
        :return: list of tuples
        """

        ret = []

        for key, val in getattr(cls, '_fields').items():
            if hasattr(val, prop_key):
                ret.append((key, getattr(val, prop_key)))
        return ret

    @classmethod
    def to_exceptions(cls, errors):
        """ Convert the validation errors into ValidationFailure exc's

        Transform native schematics validation errors into a
        goldman ValidationFailure exception.

        :param errors:
            dict of errors in schematics format
        :return:
            list of ValidationFailiure exception objects
        """

        ret = []

        for key, val in errors.items():
            if key in cls.relationships:
                attr = '/data/relationships/%s' % key
            else:
                attr = '/data/attributes/%s' % key

            for error in val:
                ret.append(ValidationFailure(attr, detail=error))

        return ret

    @property
    def dirty(self):
        """ Return a boolean indicating the dirty state of the model """

        return bool(self.dirty_fields)

    @property
    def dirty_fields(self):
        """ Return an array of field names that are dirty

        Dirty means if a model was hydrated first from the
        store & then had field values changed they are now
        considered dirty.

        For new models all fields are considered dirty.

        :return: list
        """

        dirty_fields = []

        for field in self.all_fields:
            if field not in self._original:
                dirty_fields.append(field)
            elif self._original[field] != getattr(self, field):
                dirty_fields.append(field)
        return dirty_fields

    @property
    def rid_value(self):
        """ Return the resource id field value """

        return getattr(self, self.rid_field)

    @property
    def rtype_value(self):
        """ Return the resource rtype field value """

        return getattr(self, self.rtype_field)

    def merge(self, data, clean=False, validate=False):
        """ Merge a dict with the model

        This is needed because schematics doesn't auto cast
        values when assigned. This method allows us to ensure
        incoming data & existing data on a model are always
        coerced properly.

        We create a temporary model instance with just the new
        data so all the features of schematics deserialization
        are still available.

        :param data:
            dict of potentially new different data to merge

        :param clean:
            set the dirty bit back to clean. This is useful
            when the merge is coming from the store where
            the data could have been mutated & the new merged
            in data is now the single source of truth.

        :param validate:
            run the schematics validate method

        :return:
            nothing.. it has mutation side effects
        """

        try:
            model = self.__class__(data)
        except ConversionError as errors:
            abort(self.to_exceptions(errors.messages))

        for key, val in model.to_native().items():
            if key in data:
                setattr(self, key, val)

        if validate:
            try:
                self.validate()
            except ModelValidationError as errors:
                abort(self.to_exceptions(errors.messages))

        if clean:
            self._original = self.to_native()

    def to_primitive(self, load_rels=None, sparse_fields=None, *args,
                     **kwargs):
        """ Override the schematics native to_primitive method

        :param loads_rels:
            List of field names that are relationships that should
            be loaded for the serialization process. This needs
            to be run before the native schematics to_primitive is
            run so the proper data is serialized.

        :param sparse_fields:
            List of field names that can be provided which limits
            the serialization to ONLY those field names. A whitelist
            effectively.
        """

        if load_rels:
            for rel in load_rels:
                getattr(self, rel).load()

        data = super(Model, self).to_primitive(*args, **kwargs)

        if sparse_fields:
            for key in data.keys():
                if key not in sparse_fields:
                    del data[key]

        return data



"""
    extensions
    ~~~~~~~~~~

    Define asome constants for different mimetypes
"""


EXTENSION_TABLE = {
    'image/gif': 'gif',
    'image/jpeg': 'jpeg',
    'image/jpg': 'jpg',
    'image/png': 'png',
}


def get(content_type):
    """ Return the string extension from content-type """

    return EXTENSION_TABLE.get(content_type)



"""
    types.longitude
    ~~~~~~~~~~~~~~~

    schematics longitude type
"""

from schematics.types import DecimalType


class Type(DecimalType):
    """ Longitude field with validation """

    def __init__(self, **kwargs):

        kwargs.update({
            'min_value': -180,
            'max_value': 180,
        })

        super(Type, self).__init__(**kwargs)



"""
    deserializers.base
    ~~~~~~~~~~~~~~~~~~

    Our base deserializer to be inherited by all other
    deserializers.

    Each deserializer that sub-classes this object should
    set a class constant of MIMETYPE so the deserializer is
    auto-selected during content negotiation.
"""

import goldman.exceptions as exceptions

from goldman.utils.error_helpers import abort


class Deserializer(object):
    """ Our base deserializer for sub-classing """

    MIMETYPE = ''

    def __init__(self, req, resp):

        self.req = req
        self.resp = resp

    def deserialize(self):
        """ Invoke the deserializer

        The base deserializer may eventually need to do some
        common stuff that would be annoying for each & every
        deserializer to implement.
        """

        pass

    @staticmethod
    def fail(detail, link):
        """ Convenience method aborting on non-compliant request bodies

        Call this method with a string description of the reason
        for the error & a URL. The URL is typically a section in
        an online specification citing the proper way to construct
        the request body
        """

        abort(exceptions.InvalidRequestBody(**{
            'detail': detail,
            'links': link
        }))



"""
    middleware
    ~~~~~~~~~~

    Our custom middlewares as documented here:

        http://falcon.readthedocs.org/en/stable/api/middleware.html

    Things like auth, logging, analytics, etc can go here.
"""

from ..middleware.basicauth import Middleware as BasicAuthMiddleware
from ..middleware.bearer_token import Middleware as BearerTokenMiddleware
from ..middleware.deserializer import Middleware as DeserializerMiddleware
from ..middleware.http_specs import Middleware as HttpSpecsMiddleware
from ..middleware.model_qps import Middleware as ModelQpsMiddleware
from ..middleware.rate_limit import Middleware as RateLimitMiddleware
from ..middleware.security import Middleware as SecurityMiddleware
from ..middleware.serializer import Middleware as SerializerMiddleware
from ..middleware.threadlocal import Middleware as ThreadLocalMiddleware


MIDDLEWARES = [
    BasicAuthMiddleware,
    BearerTokenMiddleware,
    DeserializerMiddleware,
    HttpSpecsMiddleware,
    ModelQpsMiddleware,
    RateLimitMiddleware,
    SecurityMiddleware,
    SerializerMiddleware,
    ThreadLocalMiddleware,
]



"""
    queryparams.filter
    ~~~~~~~~~~~~~~~~~~

    Filter resources according to one or more criteria.

    Like django & other frameworks the key & the operator are
    separated by a double underscore. Example filters:

        filter[name]=John
        filter[price__gt]=199
        filter[boss__exists]=true
        filter[loc__geo_near]=40,90,1


    The query parameter itself follows the JSON API convention
    for filtering:

        jsonapi.org/format/#fetching-filtering
        jsonapi.org/recommendations/#filtering
"""

import goldman
import re

from goldman.exceptions import InvalidQueryParams
from goldman.utils.str_helpers import str_to_bool, str_to_dt


LINK = 'jsonapi.org/format/#fetching-filtering'
PARAM = 'filter'


class Filter(object):
    """ Filter parameter object

    An instance of this object can be used to express a
    filter critera for later use by the store or whatever.

    :param field:
        string field name to filter on
    :param oper:
        string operator expression
    :param val:
        Value for expression evaluation & can be any data
        type that is expected by the store. It should already
        have been cast into whatever data type is expected.
    """

    def __init__(self, field, oper, val):

        self.field = field
        self.oper = oper
        self.val = val

    def __eq__(self, other):
        """ Compare other Filter objects """

        try:
            return self.field == other.field and \
                self.oper == other.oper and \
                self.val == other.val
        except AttributeError:
            return False

    def __repr__(self):

        name = self.__class__.__name__
        return '%s(\'%s\', \'%s\', \'%s\')' % (name, self.field, self.oper,
                                               self.val)

    def __str__(self):

        return 'filter[%s__%s]=%s' % (self.field, self.oper, self.val)


class FilterOr(object):
    """ OR'd filter expression

    This object will represent the collection of Filter
    objects that should be evaluated as a collection using
    the logical `OR` operator.
    """

    def __init__(self, filters):

        self.filters = filters

    def __contains__(self, item):
        """ Leverage the Filter __eq__ for comparison logic """

        return item in self.filters

    def __getitem__(self, index):

        return self.filters[index]

    def __len__(self):

        return len(self.filters)

    def __repr__(self):

        name = self.__class__.__name__
        return '%s(%s)' % (name, self.filters)


class FilterRel(Filter):
    """ Relationship filter expression

    This simply sub-classes the Filter object & adds some
    additional helper attributes which can later be used to
    generate queries.

    An example filter query of:
        filter[creator.username]=john

    would have the following `FilterRel` attribute values:

        local_field = creator
        foreign_filter = username
        foreign_field = rid
        foreign_rtype = logins

    foreign_field & foreign_rtype typically represent the
    same named properties on our goldman relationship types.
    """

    def __init__(self, foreign_field, foreign_filter, foreign_rtype,
                 local_field, *args, **kwargs):

        super(FilterRel, self).__init__(*args, **kwargs)

        self.foreign_filter = foreign_filter
        self.foreign_field = foreign_field
        self.foreign_rtype = foreign_rtype
        self.local_field = local_field


def _parse_param(key):
    """ Parse the query param looking for filters

    Determine the field to filter on & the operator to
    be used when filtering.

    :param key:
        The query parameter to the left of the equal sign
    :return:
        tuple of string field name & string operator
    """

    regex = re.compile(r'filter\[([A-Za-z0-9_.]+)\]')
    match = regex.match(key)

    if match:
        field_and_oper = match.groups()[0].split('__')

        if len(field_and_oper) == 1:
            return field_and_oper[0], 'eq'
        elif len(field_and_oper) == 2:
            return tuple(field_and_oper)
        else:
            raise InvalidQueryParams(**{
                'detail': 'The filter query param of "%s" is not '
                          'supported. Multiple filter operators are '
                          'not allowed in a single expression.' % key,
                'links': LINK,
                'parameter': PARAM,
            })


def _validate_field(param, fields):
    """ Ensure the field exists on the model """

    if '.' not in param.field and param.field not in fields:
        raise InvalidQueryParams(**{
            'detail': 'The filter query param of "%s" is not possible. The '
                      'resource requested does not have a "%s" field. Please '
                      'modify your request & retry.' % (param, param.field),
            'links': LINK,
            'parameter': PARAM,
        })


def _validate_rel(param, rels):
    """ Validate relationship based filters

    We don't support nested filters currently.

    FIX: Ensure the relationship filter field exists on the
         relationships model!
    """

    if param.field.count('.') > 1:
        raise InvalidQueryParams(**{
            'detail': 'The filter query param of "%s" is attempting to '
                      'filter on a nested relationship which is not '
                      'currently supported.' % param,
            'links': LINK,
            'parameter': PARAM,
        })
    elif '.' in param.field:
        model_field = param.field.split('.')[0]

        if model_field not in rels:
            raise InvalidQueryParams(**{
                'detail': 'The filter query param of "%s" is attempting to '
                          'filter on a relationship but the "%s" field is '
                          'NOT a relationship field.' % (param, model_field),
                'links': LINK,
                'parameter': PARAM,
            })


def _validate_param(param):  # pylint: disable=too-many-branches
    """ Ensure the filter cast properly according to the operator """

    detail = None

    if param.oper not in goldman.config.QUERY_FILTERS:
        detail = 'The query filter {} is not a supported ' \
                 'operator. Please change {} & retry your ' \
                 'request'.format(param.oper, param)

    elif param.oper in goldman.config.GEO_FILTERS:
        try:
            if not isinstance(param.val, list) or len(param.val) <= 2:
                raise ValueError
            else:
                param.val = [float(i) for i in param.val]
        except ValueError:
            detail = 'The query filter {} requires a list ' \
                     'of floats for geo evaluation. Please ' \
                     'modify your request & retry'.format(param)

    elif param.oper in goldman.config.ENUM_FILTERS:
        if not isinstance(param.val, list):
            param.val = [param.val]

        param.val = tuple(param.val)

    elif isinstance(param.val, list):
        detail = 'The query filter {} should not be specified more ' \
                 'than once or have multiple values. Please modify ' \
                 'your request & retry'.format(param)

    elif param.oper in goldman.config.BOOL_FILTERS:
        try:
            param.val = str_to_bool(param.val)
        except ValueError:
            detail = 'The query filter {} requires a boolean ' \
                     'for evaluation. Please modify your ' \
                     'request & retry'.format(param)

    elif param.oper in goldman.config.DATE_FILTERS:
        try:
            param.val = str_to_dt(param.val)
        except ValueError:
            detail = 'The query filter {} supports only an ' \
                     'epoch or ISO 8601 timestamp. Please ' \
                     'modify your request & retry'.format(param)

    elif param.oper in goldman.config.NUM_FILTERS:
        try:
            param.val = int(param.val)
        except ValueError:
            detail = 'The query filter {} requires a number ' \
                     'for evaluation. Please modify your ' \
                     'request & retry'.format(param)

    if detail:
        raise InvalidQueryParams(**{
            'detail': detail,
            'links': LINK,
            'parameter': PARAM,
        })


def init(req, model):
    """ Return an array of Filter objects. """

    fields = model.all_fields
    rels = model.relationships
    params = []

    for key, val in req.params.items():
        try:
            field, oper = _parse_param(key)
        except ValueError:
            continue

        try:
            local_field, foreign_filter = field.split('.')
            field_type = getattr(model, local_field)

            foreign_field = field_type.field
            foreign_rtype = field_type.rtype

            if hasattr(field_type, 'local_field'):
                local_field = field_type.local_field

            param = FilterRel(foreign_field, foreign_filter, foreign_rtype,
                              local_field, field, oper, val)
        except AttributeError:
            raise InvalidQueryParams(**{
                'detail': 'The filter query param "%s" specified a filter '
                          'containing a "." indicating a relationship filter '
                          'but a relationship by that name does not exist '
                          'on the requested resource.' % key,
                'links': LINK,
                'parameter': PARAM,
            })
        except ValueError:
            param = Filter(field, oper, val)
        else:
            _validate_param(param)
            _validate_rel(param, rels)
            _validate_field(param, fields)
            params.append(param)

    return params



"""
    resources.oauth_revoke
    ~~~~~~~~~~~~~~~~~~~~~~

    OAuth2 Token Revocation resource object with responders.

    This resource should be used to accept access_token
    revocation requests according to RFC 7009.

    The resource requires a callable to be passed in as the
    revoke_token property which will be given a token to
    revoke. The callable's return code is ignored according
    to section 2.2 of RFC 7009.

    NOTE: currently we only support revocation of access_token
          types since goldman doesn't support refresh_tokens yet
"""

import falcon
import goldman

from ..resources.base import Resource as BaseResource


class Resource(BaseResource):
    """ OAuth2 Token Revocation resource """

    DESERIALIZERS = [
        goldman.FormUrlEncodedDeserializer,
    ]

    SERIALIZERS = [
        goldman.JsonSerializer,
    ]

    def __init__(self, revoke_token):

        self.revoke_token = revoke_token
        super(Resource, self).__init__()

    def on_post(self, req, resp):
        """ Validate the token revocation request for spec compliance

        The spec also dictates the JSON based error response
        on failure & is handled in this responder.
        """

        token = req.get_param('token')
        token_type_hint = req.get_param('token_type_hint')

        # errors or not, disable client caching along the way
        # per the spec
        resp.disable_caching()

        if not token:
            resp.status = falcon.HTTP_400
            resp.serialize({
                'error': 'invalid_request',
                'error_description': 'A token parameter is required during '
                                     'revocation according to RFC 7009.',
                'error_uri': 'tools.ietf.org/html/rfc7009#section-2.1',
            })
        elif token_type_hint == 'refresh_token':
            resp.status = falcon.HTTP_400
            resp.serialize({
                'error': 'unsupported_token_type',
                'error_description': 'Currently only access_token types can '
                                     'be revoked, NOT refresh_token types.',
                'error_uri': 'tools.ietf.org/html/rfc7009#section-2.2.1',
            })
        else:
            # ignore return code per section 2.2
            self.revoke_token(token)
            resp.status = falcon.HTTP_200



"""
    types.state
    ~~~~~~~~~~~

    schematics united states, state type
"""

import us

from schematics.exceptions import ConversionError
from schematics.types import BaseType


class Type(BaseType):
    """ U.S. state field with validation """

    MESSAGES = {
        'convert': 'invalid U.S. state'
    }

    def to_native(self, value, context=None):
        """ Schematics deserializer override

        We return a us.states.State object so the abbreviation
        or long name can be trivially accessed. Additionally,
        some geo type ops are available.

        TIP: The us.states library expects unicode so we
             must ensure value is encoded properly.

        :return: us.states.State
        """

        if isinstance(value, us.states.State):
            return value

        try:
            state = us.states.lookup(unicode(value))
            if not state:
                raise TypeError
            return state
        except TypeError:
            raise ConversionError(self.messages['convert'])

    def to_primitive(self, value, context=None):
        """ Schematics serializer override

        By default we convert to a str & use the states long name.

        :return: str
        """

        if value:
            return str(value)
        return value



"""
    serializers.jsonapi
    ~~~~~~~~~~~~~~~~~~~

    Serializer that is compliant with the JSON API specification
    as documented here:

        http://jsonapi.org/
"""

import goldman
import json

from goldman.utils.url_helpers import rid_url
from ..serializers.base import Serializer as BaseSerializer


class Serializer(BaseSerializer):
    """ JSON API compliant serializer """

    MIMETYPE = goldman.JSONAPI_MIMETYPE

    def serialize(self, data):
        """ Determine & invoke the proper serializer method

        If data is a list then the serialize_datas method will
        be run otherwise serialize_data.
        """

        super(Serializer, self).serialize(data)

        body = {
            'jsonapi': {
                'version': goldman.config.JSONAPI_VERSION,
            },
            'links': {
                'self': self.req.path,
            },
        }

        included = data['included']
        if included:
            body['included'] = self._serialize_datas(included)

        _data = data['data']
        if isinstance(_data, list):
            body.update({'data': self._serialize_datas(_data)})
            body.update({'links': self._serialize_pages()})
        elif _data:
            body.update({'data': self._serialize_data(_data)})
        else:
            body.update({'data': None})

        self.resp.body = json.dumps(body)

    def _serialize_datas(self, datas):
        """ Turn the list into JSON API compliant resource objects

        :spec:
            jsonapi.org/format/#document-top-level
        :param datas:
            list of dicts
        :return:
            list resources in JSON API format
        """

        return [self._serialize_data(data) for data in datas]

    def _serialize_data(self, data):
        """ Turn the data into a JSON API compliant resource object

        WARN: This function has both side effects & a return.
              It's complete shit because it mutates data &
              yet returns a new doc. FIX.

        :spec:
            jsonapi.org/format/#document-resource-objects
        :param data:
            dict for serializing
        :return:
            dict resource in JSON API format
        """

        rels = {}
        rlink = rid_url(data['rtype'], data['rid'])

        doc = {
            'id': data.pop('rid'),
            'type': data.pop('rtype'),
            'links': {
                'self': rlink,
            },
        }

        for key, val in data['to_many'].items():
            rels.update(self._serialize_to_many(key, val, rlink))
        del data['to_many']

        for key, val in data['to_one'].items():
            rels.update(self._serialize_to_one(key, val, rlink))
        del data['to_one']

        if data:
            doc['attributes'] = data
        if rels:
            doc['relationships'] = rels

        return doc

    def _serialize_pages(self):
        """ Return a JSON API compliant pagination links section """

        path = self.req.path
        pages = self.req.pages
        links = {
            'self': '%s?%s' % (path, pages.current),
            'first': None,
            'last': None,
            'next': None,
            'prev': None,
        }

        first = pages.first
        if first:
            links['first'] = '%s?%s' % (path, first)

        last = pages.last
        if last:
            links['last'] = '%s?%s' % (path, last)

        more = pages.more
        if more:
            links['next'] = '%s?%s' % (path, more)

        prev = pages.prev
        if prev:
            links['prev'] = '%s?%s' % (path, prev)

        return links

    def _serialize_to_many(self, key, vals, rlink):
        """ Make a to_many JSON API compliant

        :spec:
            jsonapi.org/format/#document-resource-object-relationships
        :param key:
            the string name of the relationship field
        :param vals:
            array of dict's containing `rid` & `rtype` keys for the
            to_many, empty array if no values, & None if the to_manys
            values are unknown
        :return:
            dict as documented in the spec link
        """

        rel = {
            key: {
                'data': [],
                'links': {
                    'related': rlink + '/' + key
                }
            }
        }

        try:
            for val in vals:
                rel[key]['data'].append({
                    'id': val['rid'],
                    'type': val['rtype'],
                })
        except TypeError:
            del rel[key]['data']

        return rel

    def _serialize_to_one(self, key, val, rlink):
        """ Make a to_one JSON API compliant

        :spec:
            jsonapi.org/format/#document-resource-object-relationships
        :param key:
            the string name of the relationship field
        :param val:
            dict containing `rid` & `rtype` keys for the to_one &
            None if the to_one is null
        :return:
            dict as documented in the spec link
        """

        data = None
        if val and val['rid']:
            data = {'id': val['rid'], 'type': val['rtype']}

        return {
            key: {
                'data': data,
                'links': {
                    'related': rlink + '/' + key
                }
            }
        }



"""
    middleware.threadlocal
    ~~~~~~~~~~~~~~~~~~~~~~

    This middleware will initialize the native python thread.local
    capability common in other python web frameworks.

    Any attributes anchored on the goldman.sess attribute is
    unique & isolated to the thread handling the request.
"""

import goldman


class Middleware(object):
    """ Thread local storage middleware. """

    # pylint: disable=unused-argument
    def process_request(self, req, resp):
        """ Process the request before routing it. """

        goldman.sess.req = req

        if goldman.config.STORE:
            goldman.sess.store = goldman.config.STORE()



"""
    utils.error_helpers
    ~~~~~~~~~~~~~~~~~~~

    Convenient interfaces for aborting on error similar
    to abort() in the Flask micro-web framework.
"""

from goldman.exceptions import AccessDenied, ModificationDenied
from goldman.serializers.jsonapi_error import Serializer as \
    JsonApiErrorSerializer


__all__ = ['abort', 'access_fail', 'mod_fail']


def abort(error):
    """ Immediately raise our serializer with error(s)

    See jsonapi_error.Serializer docs on the usage of error
    """

    raise JsonApiErrorSerializer(error)


def access_fail(msg=None):
    """ Simple wrapper around aborting with AccessDenied """

    if msg:
        abort(AccessDenied(**{
            'detail': msg,
        }))
    else:
        abort(AccessDenied)


def mod_fail(msg=None):
    """ Simple wrapper around aborting with ModificationDenied """

    if msg:
        abort(ModificationDenied(**{
            'detail': msg,
        }))
    else:
        abort(ModificationDenied)



"""
    types.latitude
    ~~~~~~~~~~~~~~

    schematics latitude type
"""

from schematics.types import DecimalType


class Type(DecimalType):
    """ Latitude field with validation """

    def __init__(self, **kwargs):

        kwargs.update({
            'min_value': -90,
            'max_value': 90,
        })

        super(Type, self).__init__(**kwargs)



"""
    types.phonenumber
    ~~~~~~~~~~~~~~~~~

    schematics united states phone number type
"""

import phonenumbers as pn

from phonenumbers.phonenumberutil import NumberParseException
from schematics.exceptions import ConversionError
from schematics.types import BaseType


class Type(BaseType):
    """ U.S. phone number validation """

    MESSAGES = {
        'convert': 'incorrect phone number format',
        'invalid': 'not a reachable U.S. phone number',
    }

    def to_native(self, value, context=None):
        """ Schematics deserializer override

        We return a phoenumbers.PhoneNumber object so any kind
        of formatting can be trivially performed. Additionally,
        some convenient properties have been added:

            e164: string formatted '+11234567890'
            pretty: string formatted '(123) 456-7890'

        :return: phonenumbers.PhoneNumber
        """

        if isinstance(value, pn.phonenumber.PhoneNumber):
            return value

        try:
            phone = pn.parse(value, 'US')
            valid = pn.is_valid_number(phone)
        except (NumberParseException, TypeError):
            raise ConversionError(self.messages['convert'])

        if not valid and pn.is_possible_number(phone):
            raise ConversionError(self.messages['invalid'])
        elif not valid:
            raise ConversionError(self.messages['convert'])

        phone.e164 = pn.format_number(phone, pn.PhoneNumberFormat.E164)
        phone.pretty = pn.format_number(phone, pn.PhoneNumberFormat.NATIONAL)

        return phone

    def to_primitive(self, value, context=None):
        """ Schematics serializer override

        By default we convert to a str & pretty print the phone
        number.

        :return: str
        """

        if hasattr(value, 'pretty'):
            return value.pretty
        return value



"""
    serializers.base
    ~~~~~~~~~~~~~~~~

    Our base serializer to be sub-classed by the other
    serializers.
"""


class Serializer(object):
    """ Our base serializer for sub-classing """

    MIMETYPE = ''

    def __init__(self, req, resp):

        self.req = req
        self.resp = resp

    # pylint: disable=unused-argument
    def serialize(self, data):
        """ Invoke the serializer

        These are common things for all serializers. Mostly,
        stuff to do with managing headers. The data passed
        in may not be reliable for much of anything.

        Conditionally, set the Content-Type header unless it
        has already been set.
        """

        if not self.resp.content_type:
            self.resp.set_header('Content-Type', getattr(self, 'MIMETYPE'))



"""
    deserializers.json_7159
    ~~~~~~~~~~~~~~~~~~~~~~~

    Deserializer that is compliant with RFC 7159 (JSON spec).

    To avoid name collisions with the python json module we
    name ours json_7159.
"""

import goldman
import json

from ..deserializers.base import Deserializer as BaseDeserializer


class Deserializer(BaseDeserializer):
    """ JSON compliant deserializer """

    MIMETYPE = goldman.JSON_MIMETYPE

    def deserialize(self):
        """ Invoke the RFC 7159 spec compliant parser

        :return:
            the parsed & vetted request body
        """

        super(Deserializer, self).deserialize()

        try:
            return json.loads(self.req.get_body())
        except TypeError:
            link = 'tools.ietf.org/html/rfc7159'
            self.fail('Typically, this error is due to a missing JSON '
                      'payload in your request when one was required. '
                      'Otherwise, it could be a bug in our API.', link)
        except UnicodeDecodeError:
            link = 'tools.ietf.org/html/rfc7159#section-8.1'
            self.fail('We failed to process your JSON payload & it is '
                      'most likely due to non UTF-8 encoded characters '
                      'in your JSON.', link)
        except ValueError as exc:
            link = 'tools.ietf.org/html/rfc7159'
            self.fail('The JSON payload appears to be malformed & we '
                      'failed to process it. The error with line & column '
                      'numbers is: %s' % exc.message, link)



"""
    response
    ~~~~~~~~

    Define our sub-classed Response object to be used instead
    of the native falcon Response object.
"""

from falcon.response import Response as FalconResponse


class Response(FalconResponse):
    """ Subclass the default falcon response object """

    def __init__(self, *args, **kwargs):

        super(Response, self).__init__(*args, **kwargs)

        self.serializer = None
        self._init_default_headers()

    def _init_default_headers(self):
        """ Initialize the response object with default headers

        A summary of the rational behind adding the headers
        by default on a case-by-case basis is detailed below.

        Vary Header
        ~~~~~~~~~~~

        The `Vary` header will let caches know which request
        headers should be considered when looking up a cached
        document.

        The `Vary` header should specify `Accept` since our
        resources commonly support multiple representations
        of the data (serializers). The representation is
        determined by the Accept header.

        The `Vary` header should specify `Prefer` according
        to RFC 7240.
        """

        self.set_header('Vary', 'Accept, Prefer')

    def disable_caching(self):
        """ Add some headers so the client won't cache the response

        No return code, side-effects instead.
        """

        self.set_header('Cache-Control', 'no-store')
        self.set_header('Pragma', 'no-cache')

    def serialize(self, *args, **kwargs):
        """ Simple proxy to the serializer's serialize function

        This allows code to later run response.serialize() & have it
        always call the serialize method on the proper serializer.
        """

        return self.serializer.serialize(*args, **kwargs)



"""
    models.logins
    ~~~~~~~~~~~~~

    The Login model & schema
"""

import goldman
import goldman.signals as signals

from ..models.default_schema import Model as DefaultSchemaModel
from datetime import datetime as dt
from goldman.exceptions import AuthRejected
from goldman.utils.str_helpers import (
    cmp_val_salt_hash,
    gen_salt_and_hash,
    random_str,
)
from goldman.types import DateTimeType, ResourceType
from schematics.exceptions import ValidationError
from schematics.types import BooleanType, StringType


class Model(DefaultSchemaModel):
    """ Login model """

    RTYPE = 'logins'

    """
    The attrs below are the models fields
    """

    rtype = ResourceType(RTYPE)

    # identity fields
    username = StringType(
        admin_restrict=True,
        lower=True,
        max_length=25,
        required=True,
    )

    # security fields
    locked = BooleanType(
        admin_restrict=True,
        required=True,
    )
    login_date = DateTimeType(from_rest=False)
    password = StringType(
        max_length=100,
        min_length=6,
        serialize_when_none=False,
        to_rest=False,
    )
    salt = StringType(
        from_rest=False,
        to_rest=False,
    )
    token = StringType(
        from_rest=False,
        to_rest=False,
    )

    @classmethod
    def auth_creds(cls, username, password):
        """ Callback method for Basic Authentication """

        store = goldman.sess.store
        login = store.find(cls.RTYPE, 'username', username)

        if not login:
            msg = 'No login found by that username. Spelling error?'
            raise AuthRejected(**{'detail': msg})
        elif login.locked:
            msg = 'The login account is currently locked out.'
            raise AuthRejected(**{'detail': msg})
        elif not cmp_val_salt_hash(password, login.salt, login.password):
            msg = 'The password provided is incorrect. Spelling error?'
            raise AuthRejected(**{'detail': msg})
        else:
            return login

    @classmethod
    def auth_creds_token(cls, username, password):
        """ Callback method for OAuth 2.0 ROPC resource """

        login = cls.auth_creds(username, password)

        if not login.token:
            login.token = random_str()
        return login, login.token

    @classmethod
    def auth_token(cls, token):
        """ Callback method for OAuth 2.0 bearer token middleware """

        store = goldman.sess.store
        login = store.find(cls.RTYPE, 'token', token)

        if not login:
            msg = 'No login found with that token. It may have been revoked.'
            raise AuthRejected(**{'detail': msg})
        elif login.locked:
            msg = 'The login account is currently locked out.'
            raise AuthRejected(**{'detail': msg})
        else:
            return login

    def validate_username(self, data, value):
        """ Ensure the username is unique

        If the login is being created then simply check if
        the username is in the store & fail.

        Otherwise if the login is being updated check if the
        existing rid on a username match is the same as the
        login being updated otherwise fail.
        """

        store = goldman.sess.store
        existing = store.find(data['rtype'], 'username', value)

        if existing:
            if not data['rid'] or data['rid'] != existing.rid:
                raise ValidationError('username is already taken')


# pylint: disable=unused-argument
def pre_create(sender, model):
    """ Callback before creating a new login

    Without a password during create we are forced to
    lock the account & set the password to something
    random.
    """

    if not model.password:
        model.locked, model.password = True, random_str()


def pre_save(sender, model):
    """ Hash the password if being changed """

    if 'password' in model.dirty_fields:
        model.salt, model.password = gen_salt_and_hash(model.password)


def post_authenticate(sender):
    """ Update the login_date timestamp

    The login_date update will be debounced so writes don't
    occur on every hit of the the API. If the login_date
    was modified within 15 minutes then don't update it.
    """

    login = goldman.sess.login
    now = dt.now()

    if not login.login_date:
        login.login_date = now
    else:
        sec_since_updated = (now - login.login_date).seconds
        min_since_updated = sec_since_updated / 60

        if min_since_updated > 15:
            login.login_date = now

    if login.dirty:
        store = goldman.sess.store
        store.update(login)


signals.pre_create.connect(pre_create, sender=Model)
signals.pre_save.connect(pre_save, sender=Model)

signals.post_authenticate.connect(post_authenticate)



"""
    resources.related
    ~~~~~~~~~~~~~~~~~

    Related resource object(s) with responders.

    Get the to-one or to-many relationship by name (related)
    of the model.
"""

import goldman
import goldman.signals as signals

from ..resources.base import Resource as BaseResource
from goldman.exceptions import InvalidURL
from goldman.utils.error_helpers import abort
from goldman.utils.responder_helpers import (
    find,
    to_rest_model,
    to_rest_models,
)


class Resource(BaseResource):
    """ Related item(s) resource & responders """

    DESERIALIZERS = [
        goldman.JsonApiDeserializer,
    ]

    SERIALIZERS = [
        goldman.CsvSerializer,
        goldman.JsonApiSerializer,
    ]

    def __init__(self, model):

        self.model = model
        self.rtype = model.RTYPE

        super(Resource, self).__init__()

    def on_get(self, req, resp, rid, related):
        """ Find the related model & serialize it back

        If the parent resource of the related model doesn't
        exist then abort on a 404.
        """

        signals.responder_pre_any.send(self.model)
        signals.responder_pre_find.send(self.model)

        if not hasattr(self.model, related):
            abort(InvalidURL(**{
                'detail': 'The "%s" resource does not have a related '
                          'resource named "%s". This is an error, check '
                          'your spelling & retry.' % (self.rtype, related)
            }))

        model = find(self.model, rid)
        model_related = getattr(model, related).load()

        if isinstance(model_related, list):
            props = to_rest_models(model_related, includes=req.includes)
        elif model:
            props = to_rest_model(model_related, includes=req.includes)
        else:
            props = model_related

        resp.serialize(props)

        signals.responder_post_any.send(self.model)
        signals.responder_post_find.send(self.model)



"""
    queryparams.include
    ~~~~~~~~~~~~~~~~~~~

    Determine relationship resources to include in the response
    according to one or more criteria. Documented here:

        jsonapi.org/format/#fetching-includes
"""

from goldman.exceptions import InvalidQueryParams


LINK = 'jsonapi.org/format/#fetching-includes'
PARAM = 'include'


def _validate_no_nesting(param):
    """ Ensure the include field is not a nested relationship """

    if '.' in param:
        raise InvalidQueryParams(**{
            'detail': 'The include query param of the "%s" field '
                      'is not supported. Nested relationship '
                      'inclusions are not currently supported' % param,
            'links': LINK,
            'parameter': PARAM,
        })


def _validate_rels(param, rels):
    """ Ensure the include field is a relationship """

    if param not in rels:
        raise InvalidQueryParams(**{
            'detail': 'The include query param of the "%s" field '
                      'is not possible. It does not represent a '
                      'relationship field & on the primary resource '
                      '& is not eligible for inclusion as a compound '
                      'document.' % param,
            'links': LINK,
            'parameter': PARAM,
        })


def init(req, model):
    """ Return an array of fields to include. """

    rels = model.relationships
    params = req.get_param_as_list('include') or []
    params = [param.lower() for param in params]

    for param in params:
        _validate_no_nesting(param)
        _validate_rels(param, rels)

    return params



"""
    serializers.json_7159
    ~~~~~~~~~~~~~~~~~~~~~

    Serializer that is compliant with RFC 7159 (JSON spec). To avoid
    name collisions with the python json module we name ours
    json_7159.
"""

import goldman
import json

from ..serializers.base import Serializer as BaseSerializer


class Serializer(BaseSerializer):
    """ JSON compliant serializer """

    MIMETYPE = goldman.JSON_MIMETYPE

    def serialize(self, data):
        """ Call json.dumps & let it rip """

        super(Serializer, self).serialize(data)
        self.resp.body = json.dumps(data)



"""
    types
    ~~~~~

    All of our custom schematics model types.
"""

from ..types.datetime import Type as DateTimeType
from ..types.latitude import Type as LatitudeType
from ..types.longitude import Type as LongitudeType
from ..types.phonenumber import Type as PhoneNumberType
from ..types.resource import Type as ResourceType
from ..types.state import Type as StateType
from ..types.to_many import Type as ToManyType
from ..types.to_one import Type as ToOneType
from ..types.zipcode import Type as ZipCodeType


TYPES = [
    DateTimeType,
    LatitudeType,
    LongitudeType,
    PhoneNumberType,
    ResourceType,
    StateType,
    ToManyType,
    ToOneType,
    ZipCodeType,
]



"""
    utils.responder_helpers
    ~~~~~~~~~~~~~~~~~~~~~~~

    Common helpers for our resource responders

    These can be used on any resource responders that need
    common funtionality when interacting with models in a
    RESTFUL like fashion.

    Things like URL validations, some store method wrapping,
    routine over-the-wire (de)serialization helpers, etc.
"""

import goldman
import goldman.exceptions as exceptions

from goldman.utils.error_helpers import abort, mod_fail
from schematics.types import IntType


__all__ = ['find', 'from_rest', 'to_rest_model', 'to_rest_models']


def validate_rid(model, rid):
    """ Ensure the resource id is proper """

    rid_field = getattr(model, model.rid_field)

    if isinstance(rid_field, IntType):
        try:
            int(rid)
        except (TypeError, ValueError):
            abort(exceptions.InvalidURL(**{
                'detail': 'The resource id {} in your request is not '
                          'syntactically correct. Only numeric type '
                          'resource id\'s are allowed'.format(rid)
            }))


def find(model, rid):
    """ Find a model from the store by resource id """

    validate_rid(model, rid)

    rid_field = model.rid_field
    model = goldman.sess.store.find(model.RTYPE, rid_field, rid)

    if not model:
        abort(exceptions.DocumentNotFound)

    return model


def _from_rest_hide(model, props):
    """ Purge fields not allowed during a REST deserialization

    This is done on fields with `from_rest=False`.
    """

    hide = model.get_fields_by_prop('from_rest', False)

    for field in hide:
        try:
            del props[field]
        except KeyError:
            continue


def _from_rest_ignore(model, props):
    """ Purge fields that are completely unknown """

    fields = model.all_fields

    for prop in props.keys():
        if prop not in fields:
            del props[prop]


def _from_rest_lower(model, props):
    """ Lowercase fields requesting it during a REST deserialization """

    for field in model.to_lower:
        try:
            props[field] = props[field].lower()
        except (AttributeError, KeyError):
            continue


def _from_rest_on_create(model, props):
    """ Assign the default values when creating a model

    This is done on fields with `on_create=<value>`.
    """

    fields = model.get_fields_with_prop('on_create')

    for field in fields:
        props[field[0]] = field[1]


def _from_rest_on_update(model, props):
    """ Assign the default values when updating a model

    This is done on fields with `on_update=<value>`.
    """

    fields = model.get_fields_with_prop('on_update')

    for field in fields:
        props[field[0]] = field[1]


def _from_rest_reject_update(model):
    """ Reject any field updates not allowed on POST

    This is done on fields with `reject_update=True`.
    """

    dirty = model.dirty_fields
    fields = model.get_fields_by_prop('reject_update', True)
    reject = []

    for field in fields:
        if field in dirty:
            reject.append(field)

    if reject:
        mod_fail('These fields cannot be updated: %s' % ', '.join(reject))


def from_rest(model, props):
    """ Map the REST data onto the model

    Additionally, perform the following tasks:

        * purge all fields not allowed as incoming data
        * purge all unknown fields from the incoming data
        * lowercase certain fields that need it
        * merge new data with existing & validate
            * mutate the existing model
            * abort on validation errors
            * coerce all the values
    """

    req = goldman.sess.req

    _from_rest_hide(model, props)
    _from_rest_ignore(model, props)
    _from_rest_lower(model, props)

    if req.is_posting:
        _from_rest_on_create(model, props)
    elif req.is_patching:
        _from_rest_on_update(model, props)

    model.merge(props, validate=True)

    if req.is_patching:
        _from_rest_reject_update(model)


def _to_rest_hide(model, props):
    """ Purge fields not allowed during a REST serialization

    This is done on fields with `to_rest=False`.
    """

    hide = model.get_fields_by_prop('to_rest', False)

    for field in hide:
        try:
            del props[field]
        except KeyError:
            continue


def _to_rest_includes(models, includes):
    """ Fetch the models to be included

    The includes should follow a few basic rules:

        * the include MUST not already be an array member
          of the included array (no dupes)

        * the include MUST not be the same as the primary
          data if the primary data is a single resource
          object (no dupes)

        * the include MUST not be an array member of the
          primary data if the primary data an array of
          resource objects (no dupes)

    Basically, each included array member should be the only
    instance of that resource object in the entire restified
    data.
    """

    included = []
    includes = includes or []

    if not isinstance(models, list):
        models = [models]

    for include in includes:
        for model in models:
            rel = getattr(model, include)

            if hasattr(rel, 'model') and rel.model:
                rel_models = [rel.model]
            elif hasattr(rel, 'models') and rel.models:
                rel_models = rel.models

            for rel_model in rel_models:
                if rel_model in models or rel_model in included:
                    continue
                else:
                    included.append(rel_model)

    for idx, val in enumerate(included):
        included[idx] = _to_rest(val)

    return included


def _to_rest_rels(model, props):
    """ Move the relationships to appropriate location in the props

    All to_ones should be in a to_one key while all to_manys
    should be in a to_many key.
    """

    props['to_many'] = {}
    props['to_one'] = {}

    for key in model.to_one:
        try:
            props['to_one'][key] = props.pop(key)
        except KeyError:
            continue

    for key in model.to_many:
        try:
            props['to_many'][key] = props.pop(key)
        except KeyError:
            continue


def _to_rest(model, includes=None):
    """ Convert the model into a dict for serialization

    Notify schematics of the sparse fields requested while
    also forcing the resource id & resource type fields to always
    be present no matter the request. Additionally, any includes
    are implicitly added as well & automatically loaded.

    Then normalize the includes, hide private fields, & munge
    the relationships into a format the serializers are
    expecting.
    """

    includes = includes or []
    sparse = goldman.sess.req.fields.get(model.rtype, [])

    if sparse:
        sparse += [model.rid_field, model.rtype_field]
        sparse += includes

    props = model.to_primitive(
        load_rels=includes,
        sparse_fields=sparse,
    )

    props['rid'] = props.pop(model.rid_field)
    props['rtype'] = props.pop(model.rtype_field)

    _to_rest_hide(model, props)
    _to_rest_rels(model, props)

    return props


def to_rest_model(model, includes=None):
    """ Convert the single model into a dict for serialization

    :return: dict
    """

    props = {}
    props['data'] = _to_rest(model, includes=includes)
    props['included'] = _to_rest_includes(model, includes=includes)

    return props


def to_rest_models(models, includes=None):
    """ Convert the models into a dict for serialization

    models should be an array of single model objects that
    will each be serialized.

    :return: dict
    """

    props = {}
    props['data'] = []

    for model in models:
        props['data'].append(_to_rest(model, includes=includes))

    props['included'] = _to_rest_includes(models, includes=includes)

    return props



"""
    api
    ~~~

    Our custom sub-class of the falcon.API main object.

    API will load the routes, order the middleware's, &
    register our custom request/response objects.
"""

import falcon
import goldman
import json

from goldman.request import Request
from goldman.response import Response


__all__ = ['API']


class API(falcon.API):
    """ Subclass the falcon.API object with our own

    Goldman uses custom request, response, & route loading
    techniques when compared to the falcon.API object.
    """

    MIDDLEWARE = []
    RESOURCES = []
    ROUTES = []

    def __init__(self):

        middleware = [
            goldman.SecurityMiddleware(),
            goldman.HttpSpecsMiddleware(),
            goldman.DeserializerMiddleware(),
            goldman.SerializerMiddleware(),
            goldman.ModelQpsMiddleware(),
            goldman.ThreadLocalMiddleware(),
        ]
        middleware += self.MIDDLEWARE

        super(API, self).__init__(
            middleware=middleware,
            request_type=Request,
            response_type=Response,
        )

        self._load_resources()
        self._load_routes()
        self.set_error_serializer(self._error_serializer)

    def _load_resources(self):
        """ Load all the native goldman resources.

        The route or API endpoint will be automatically determined
        based on the resource object instance passed in.

        INFO: Only our Model based resources are supported when
              auto-generating API endpoints.
        """

        for resource in self.RESOURCES:
            if isinstance(resource, goldman.ModelsResource):
                route = '/%s' % resource.rtype
            elif isinstance(resource, goldman.ModelResource):
                route = '/%s/{rid}' % resource.rtype
            elif isinstance(resource, goldman.RelatedResource):
                route = '/%s/{rid}/{related}' % resource.rtype
            else:
                raise TypeError('unsupported resource type')

            self.add_route(*(route, resource))

    def _load_routes(self):
        """ Load all the routes.

        A class constant of ROUTES is expected containing
        an array of tuples in the following format:


            ROUTES = [
                ('/<endpoint>', <resource instance>)
            ]


        A real life example with meaningful endpoints & resource
        instances would look like:


            ROUTES = [
                ('/logins', goldman.ModelsResource(LoginModel)),
                ('/logins/{rid}', goldman.ModelResource(LoginModel)),
            ]


        This is the same format the native falcon `add_route`
        method wants the routes.
        """

        for route in self.ROUTES:
            self.add_route(*route)

    # XXX FIX: API changed in falcon 0.4.0
    @staticmethod
    def _error_serializer(req, exc):  # pylint: disable=unused-argument
        """ Serializer for native falcon HTTPError exceptions.

        We override the default serializer with our own so we
        can ensure the errors are serialized in a JSON API
        compliant format.

        Surprisingly, most falcon error attributes map directly
        to the JSON API spec. The few that don't can be mapped
        accordingly:


            HTTPError               JSON API
            ~~~~~~~~~               ~~~~~~~~

            exc.description    ->   error['detail']
            exc.link['href']   ->   error['links']['about']


        Per the falcon docs this function should return a tuple
        of (MIMETYPE, BODY PAYLOAD)
        """

        error = {
            'detail': exc.description,
            'title': exc.title,
            'status': exc.status,
        }

        try:
            error['links'] = {'about': exc.link['href']}
        except (TypeError, KeyError):
            error['links'] = {'about': ''}

        return (
            goldman.config.JSONAPI_MIMETYPE,
            json.dumps({'errors': [error]}),
        )



"""
    mimetypes
    ~~~~~~~~~

    Define some constants for different mimetypes
"""


CSV_MIMETYPE = 'text/csv'
FILEUPLOAD_MIMETYPE = 'multipart/form-data'
FORMURL_MIMETYPE = 'application/x-www-form-urlencoded'
JSON_MIMETYPE = 'application/json'
JSONAPI_MIMETYPE = 'application/vnd.api+json'



"""
    deserializers.csv
    ~~~~~~~~~~~~~~~~~

    Deserializer that is compliant with the RFC 4180 CSV format

    XXX FIX: handle relationships
"""

import goldman
import goldman.exceptions as exceptions
import csv

from goldman.utils.error_helpers import abort
from ..deserializers.base import Deserializer as BaseDeserializer


class Normalizer(object):
    """ The JSON API payload normalizer """

    @classmethod
    def run(cls, row, reader):
        """ Invoke the CSV normalizer

        We don't need to vet the inputs much because the
        Parser has already done all the work.

        :return: dict
        """

        pass


class Parser(object):
    """ The CSV payload parser """

    @classmethod
    def run(cls, row, reader):
        """ Invoke the CSV parser on an individual row

        The row should already be a dict from the CSV reader.
        The reader is passed in so we can easily reference the
        CSV document headers & line number when generating
        errors.
        """

        cls._parse_keys(row, reader.line_num)
        cls._parse_relationships(row, reader.line_num)

    @staticmethod
    def _parse_keys(row, line_num):
        """ Perform some sanity checks on they keys

        Each key in the row should not be named None cause
        (that's an overrun). A key named `type` MUST be
        present on the row & have a string value.

        :param row: dict
        :param line_num: int
        """

        link = 'tools.ietf.org/html/rfc4180#section-2'

        none_keys = [key for key in row.keys() if key is None]

        if none_keys:
            fail('You have more fields defined on row number {} '
                 'than field headers in your CSV data. Please fix '
                 'your request body.'.format(line_num), link)

        elif not row.get('type'):
            fail('Row number {} does not have a type value defined. '
                 'Please fix your request body.'.format(line_num), link)

    @staticmethod
    def _parse_relationships(row, line_num):
        """ Perform some sanity checks on the relationships

        Relationships should have a single '.' in them & `id`
        & `type` keys. A string type value is required while id
        is not since it could be severing the relationship.

        :param row: dict
        :param line_num: int
        """

        pass


class Deserializer(BaseDeserializer):
    """ CSV deserializer """

    MIMETYPE = goldman.CSV_MIMETYPE

    def deserialize(self, data=None):
        """ Invoke the deserializer

        If the payload is a collection (more than 1 records)
        then a list will be returned of normalized dict's.

        If the payload is a single item then the normalized
        dict will be returned (not a list)

        :return: list or dict
        """

        data = []

        if self.req.content_type_params.get('header') != 'present':
            abort(exceptions.InvalidRequestHeader(**{
                'detail': 'When using text/csv your Content-Type '
                          'header MUST have a header=present parameter '
                          '& the payload MUST include a header of fields',
                'links': 'tools.ietf.org/html/rfc4180#section-3'
            }))

        try:
            reader = csv.DictReader(self.req.stream)

            self._validate_field_headers(reader)

            for row in reader:
                Parser.run(row, reader)

                row = Normalizer.run(row, reader)
                row = super(Deserializer, self).deserialize(data)

                data.append(row)
        except csv.Error:
            abort(exceptions.InvalidRequestBody)

        return data

    @staticmethod
    def _validate_field_headers(reader):
        """ Perform some validations on the CSV headers

        A `type` field header must be present & all field
        headers must be strings.

        :param reader: csv reader object
        """

        link = 'tools.ietf.org/html/rfc4180#section-2'

        for field in reader.fieldnames:
            if not isinstance(field, str):
                fail('All headers in your CSV payload must be '
                     'strings.', link)

        if 'type' not in reader.fieldnames:
            fail('A type header must be present in your CSV '
                 'payload.', link)



"""
    types.to_many
    ~~~~~~~~~~~~~

    schematics to_many type

    This should be used when defining a one-to-many relationship.
"""

import goldman

from goldman.queryparams.filter import Filter
from schematics.types import BaseType


class ToMany(object):
    """ ToMany object

    Evaluating whether or not to load the models from the
    store uses the is_loaded property rather than simply
    checking the models attribute. This is to needed when
    a load has been done but no models exist.
    """

    def __init__(self, rtype, field, rid):

        self.field = field
        self.rtype = rtype
        self.rid = rid

        self._is_loaded = False
        self.models = []

    def __eq__(self, other):

        try:
            return self.field == other.field and \
                self.rtype == other.rtype and \
                self.rid == other.rid
        except AttributeError:
            return False

    def __repr__(self):

        name = self.__class__.__name__

        return '%s(\'%s\', \'%s\', \'%s\')' % (name, self.rtype, self.field,
                                               self.rid)

    @property
    def is_loaded(self):
        """ Boolean indicating whether a load attempt has been made """

        return self._is_loaded

    def load(self):
        """ Return the model from the store """

        filters = [Filter(self.field, 'eq', self.rid)]
        store = goldman.sess.store

        self._is_loaded = True
        self.models = store.search(self.rtype, filters=filters)

        return self.models


class Type(BaseType):
    """ Custom field for our ToMany relationships """

    def __init__(self, deserialize_from='rid', from_rest=False, field=None,
                 rtype=None, **kwargs):

        self.local_field = deserialize_from
        self.field = field
        self.rtype = rtype

        kwargs.update({
            'deserialize_from': deserialize_from,
            'from_rest': from_rest,
        })

        super(Type, self).__init__(**kwargs)

    def to_native(self, value, context=None):
        """ Schematics deserializer override

        :return: ToMany instance
        """

        if isinstance(value, ToMany):
            return value
        else:
            return ToMany(self.rtype, self.field, value)

    def to_primitive(self, value, context=None):
        """ Schematics serializer override

        The return values vary to indicate whether or not the
        relationships is loaded with values (list with members),
        loaded but no values (empty list), or never loaded &
        unknown if values exist or not.
        """

        data = []

        if not value.is_loaded:
            return None

        for model in value.models:
            if context and context.get('rel_ids'):
                data.append(model.rid_value)
            else:
                data.append({'rtype': model.rtype, 'rid': model.rid_value})

        return data
